import os
import gtk
import time
import calendar as mod_calendar

from gettext import gettext as _

from .. import gui
from .. import config
from ..com import sushi
from ..helper import history

def strip_date(text):
	new = []
	for line in text.split("\n"):
		match = history.DATEPATTERN.match(line)
		if match:
			time_len = len(line[match.start():match.end()].split(" ")[-1])
			new.append(line[match.end()-time_len:])
		else:
			new.append(line)
	return "\n".join(new)

class HistoryDialog(object):

	def __init__(self, history_tab):
		self.current_path = ()
		self.current_file = None
		self.current_offsets = {}

		# last search result position
		self.last_search_iter = None
		self.last_result = None
		self.search_in_progress = False

		# use log API despite remotely maki connection
		self.force_remote = False

		self.builder = gtk.Builder()

		path = config.get("uifiles","dialogs")
		self.builder.add_from_file(os.path.join(path, "history.ui"))

		self.builder.connect_signals(self)

		if not self.verify_remote():
			return

		gui.mgmt.set_font(self.builder.get_object("history_view"),
						  gui.mgmt.get_font())

		# jump to current date
		(year, month, day) = time.localtime(time.time())[0:3]
		self.builder.get_object("calendar").set_properties(
			year=year, month=month-1, day=day)

		self.fill_target_tree()

		self.switch_to_target(history_tab.server.name, history_tab.name)

		self.load_current_day()

		bar = self.builder.get_object("searchbar")
		bar.autohide = False
		bar.default_buttons = False
		bar.search_button.connect("clicked", self.search)
		bar.search_entry.connect("activate", self.search)

	def verify_remote(self):
		""" ask user if he wants to try reading logs even if maki
			is connected via sushi-remote. This is useful if maki
			runs on the same host.
		"""
		if not sushi.remote:
			return True

		d = gui.builder.question_dialog(
				_("Read history anyways?"),
				_("maki is connected remotely. It's possible that "
				  "the history is not accessible. Do you want to "
				  "try anyways?"))

		id = d.run()
		d.destroy()

		if id == gtk.RESPONSE_YES:
			self.force_remote = True
			return True
		else:
			self.builder.get_object("history_dialog").destroy()
			return False


	def fill_target_tree(self):
		""" fill target tree store with server/targets """
		store = self.builder.get_object("target_tree")

		for server in history.get_available_servers(
									force_remote=self.force_remote):
			server_iter = store.append(None, (server,))

			for conv in history.get_available_conversations(server,
									force_remote=self.force_remote):
				store.append(server_iter, (conv,))

	def switch_to_target(self, server, target):
		""" switch to combo box entry identified by server / target """
		cbox = self.builder.get_object("target_combobox")
		store = self.builder.get_object("target_tree")

		for srow in store:
			if srow[0].lower() == server.lower():
				for trow in srow.iterchildren():
					if trow[0].lower() == target.lower():
						cbox.set_active_iter(trow.iter)
						d = self.builder.get_object("history_dialog")
						d.set_title(_("History for %(target)s") % {
									"target": target})

	def get_current_names(self):
		""" return (server, target) of current selection
			return (server, None) if no target is active
			return (None, None) if no entry is active
		"""
		if not self.current_path:
			return (None, None)

		store = self.builder.get_object("target_tree")
		iter = store.get_iter(self.current_path)
		target = store.get_value(iter, 0)
		parent_iter = store.iter_parent(iter)

		if not parent_iter:
			return (target, None)

		server = store.get_value(parent_iter, 0)
		return (server, target)

	def update_calendar(self, highlight=True):
		""" update the calendar markings if highlight is True and
			the cache variables self.current_file and self.current_offsets
		"""
		calendar = self.builder.get_object("calendar")
		calendar.clear_marks()

		(server, target) = self.get_current_names()

		if not server or not target:
			return

		(year, month) = calendar.get_properties("year","month")
		month += 1 # 1-12 instead of 0-11

		for log in history.get_available_logs(server, target,
										force_remote=self.force_remote):
			(lyear, lmonth) = history.get_log_date(log)

			if year == lyear and lmonth == month:
				path = history.get_log_path(server, target, log)

				try:
					fd = file(path, "r")
				except Exception,e:
					print "Exception %s while open %s." % (e, path)
					return

				self.current_file = path
				self.current_offsets = history.parse_day_offsets(fd)

				if highlight:
					for (year, month, day) in self.current_offsets.keys():
						calendar.mark_day(day)


	def search_calender_marks(self):
		needle = self.builder.get_object(
					"searchbar").search_entry.get_text()

		if not self.current_file or not needle:
			return

		calender = self.builder.get_object("calendar")
		calender.clear_marks()

		if not self.current_offsets:
			return

		fd = file(self.current_file, "r")
		for ((year, month, day),
			(start,end)) in self.current_offsets.items():

			fd.seek(start)

			if fd.read(end-start).find(needle) >= 0:
				calender.mark_day(day)


	def search_local(self):
		""" search in the text loaded into history_buffer and highlight
			the needle if found.
			Returns True on success otherwise False
		"""
		view = self.builder.get_object("history_view")
		buffer = self.builder.get_object("history_buffer")
		needle = self.builder.get_object(
					"searchbar").search_entry.get_text()

		if not needle:
			return False

		if (not self.last_search_iter or self.last_result != needle):
			# new search
			self.last_search_iter = buffer.get_start_iter()
			self.search_calender_marks()

		result = self.last_search_iter.forward_search(needle,
									gtk.TEXT_SEARCH_TEXT_ONLY)

		if not result or result == self.last_result:
			return False

		self.last_search_iter = result[1]
		self.last_result = buffer.get_text(*result)

		buffer.select_range(*result)

		# scroll the textview to the result
		view.scroll_to_iter(result[0], 0.0)
		return True

	def load_next_month(self):
		""" find the next possible month with data in it and
			jump to it. Return True on success, otherwise False
		"""
		calendar = self.builder.get_object("calendar")

		(year, month) = calendar.get_properties("year","month")

		real_month = month + 1

		(server, target) = self.get_current_names()

		if not server and not target:
			return False

		available_months = {}

		for log in history.get_available_logs(server, target,
									force_remote=self.force_remote):

			(dyear, dmonth) = history.get_log_date(log)

			if dyear > year or (dyear == year and dmonth > real_month):
				if not available_months.has_key(dyear):
					available_months[dyear] = []
				available_months[dyear].append(dmonth)

		if not available_months:
			return False

		# get lowest possible year
		low_year = sorted(available_months.keys())[0]

		# get lowest month in year
		low_month = sorted(available_months[low_year])[0] - 1

		# get first day
		# XXX this mustnt be a day which has search results so it would
		# probably better to set the day to a day with search result
		day = mod_calendar.monthrange(low_year, low_month+1)[0]

		calendar.set_properties(year=low_year, month=low_month, day=day)
		return True

	def search(self,*x):
		""" search for a needle, mark all days where something
			was found. On every hit, iterate further until no
			further matches are found.
		"""
		needle = self.builder.get_object(
					"searchbar").search_entry.get_text()
		if not needle:
			print "aborting search.."
			self.abort_search()
			return

		self.search_in_progress = True
		print "in search"

		if not self.search_local():
			# no results at current day, search for the next one.
			print "local didnt succeed..."
			calendar = self.builder.get_object("calendar")
			(cyear,cmonth,cday) = calendar.get_properties("year","month",
														  "day")
			cmonth += 1
			possible_days = []
			for (year, month, day) in self.current_offsets:
				if year == cyear and cmonth == month and day > cday:
					possible_days.append(day)
			if possible_days:
				print "got possible days, %s" % (possible_days,)
				# switch to smallest possible day
				calendar.select_day(sorted(possible_days)[0])
				self.search()
			else:
				# no days with data this month. load next month (if avail)
				if self.load_next_month():
					print "loading next month"
					return self.search()
				else:
					# no next month, abort search
					self.reset_search()
					print "SEARCH ENDED!"
					return

	def reset_search(self):
		""" reset search variables """
		self.search_in_progress = False
		self.last_search_iter = None

	def abort_search(self):
		""" clear the search markup and reset variables """
		self.reset_search()
		self.update_calendar()

	def load_current_day(self):
		calendar = self.builder.get_object("calendar")
		if not self.current_file:
			return False
		(year, month, day) = calendar.get_properties("year", "month",
													"day")
		month += 1 # we work with 1-12 not 0-11 like the calendar widget
		if not self.current_offsets.has_key((year, month, day)):
			return False

		(start, end) = self.current_offsets[(year, month, day)]

		buffer = self.builder.get_object("history_buffer")

		fd = file(self.current_file, "r")
		fd.seek(start)
		buffer.set_text(strip_date(fd.read(end - start)))
		return True

	def calendar_date_changed(self, calendar):
		if not self.search_in_progress:
			self.update_calendar()
		else:
			self.update_calendar(highlight=False)
			self.search_calender_marks()

	def calendar_day_selected(self, calendar):
		if not self.load_current_day():
			self.builder.get_object("history_buffer").set_text("")

	def target_combobox_changed(self, box):
		self.current_path = box.get_model().get_path(box.get_active_iter())
		self.update_calendar()
		self.search_in_progress = False

	def history_dialog_response(self, dialog, id):
		dialog.destroy()

	def history_buffer_changed(self, buffer, *x):
		if self.search_in_progress:
			self.last_search_iter = buffer.get_start_iter()



def run(tab):
	d = HistoryDialog(tab)

	dwin = d.builder.get_object("history_dialog")
	dwin.show_all()


def setup():
	pass



"""
Copyright (c) 2009 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import logging

from .. import builder
from .. import mgmt
from .. import shortcuts
from .._builder import widgets

from .. import _status_manager
status = _status_manager.status

from ... import config

from ...lib.input_history import InputHistory
from ...lib.nick_list_store import NickListStore

from ...helper import markup

from ...typecheck import types

# current path/tab/tabs tracking
# - get_current_path
# - get_current_tab
# - get_current_tabs
from .current import *

# make message types accessible
# XXX this should probably have it's own namespace
from .messages import *

from .tab import TekkaTab
from .server import TekkaServer
from .channel import TekkaChannel
from .query import TekkaQuery


_callbacks = {}


"""
This module provides basic access to the tabs stored in the server list.

This module also provides callbacks to get notified about changes.

List of valid callbacks:

	"new_message": (tab, msgtype)
		new message arrived on the tab

	"new_name": (tab, name)
		tab received a new name (happens often in queries)

	"add": (tab)
		this tab was added

	"remove": (tab)
		this tab is about to removed.

	"new_markup": (tab)
		tab received a new markup in the UI

	"server_connected": (tab, connected)
		connected == True -> connected
		connected == False -> disconnected

	"joined": (tab, switch)
		a change on the join state
		switch == True -> joined
		switch == False -> parted

	"new_nick": (tab, nick)
		nick changed on tab

	"tab_switched": (oldtab, newtab)
		tab switched from old to new

	"topic": (tab, topic)
		topic changed

"""


@types (d = dict)
def add_callbacks(d):
	""" get dict with {<str>:<function>,...}
		and apply the values to the internal
		callback dict as a list, resulting in:
		{<str>:[<function0>,<function1>...],...}
	"""
	global _callbacks
	for (key, value) in d.items():
		if _callbacks.has_key(key):
			_callbacks[key].append(value)
		else:
			_callbacks[key] = [value]


@types (key = basestring)
def get_callbacks(key):
	""" return all callback functions stored
		under the given key as a list.
	"""
	global _callbacks
	try:
		return _callbacks[key]
	except KeyError:
		raise ValueError, "No such signal handler: %s." % (key)


@types (callbacks = tuple)
def connect_tab_callbacks(obj, callbacks, *args):
	""" connect the object with all the callbacks
		identified by a string in the callbacks
		tuple.
	"""
	for cb in callbacks:
		cbFunctions = get_callbacks(cb)

		for fun in cbFunctions:
			obj.connect(cb, fun, *args)



def call_callback(name, *args):
	for cb in get_callbacks(name):
		cb(*args)


@types(tabtype = TekkaTab)
def _create_tab(tabtype, name, *args, **kwargs):
	""" instance class of type tabtype, connect signals,
		create output window and setup input history.

		Returns a new child of TekkaTab.
	"""
	tab = tabtype(name, *args, **kwargs)

	tab.window = builder.get_new_output_window()
	tab.window.show_all()

	mgmt.set_font(tab.window.textview, mgmt.get_font())

	connect_tab_callbacks(tab, ("new_message","new_name",
		"server_connected","new_markup"))

	tab.input_history = InputHistory(
		text_callback = widgets.get_object("input_entry").get_text)

	return tab


@types (server = TekkaServer, name = basestring)
def create_channel(server, name):
	""" create TekkaChannel object and associated a NickListStore
		with it.

		Returns the newly created Tab object.
	"""
	ns = NickListStore()

	ns.set_modes(server.support_prefix[1])

	tab = _create_tab(TekkaChannel, name, server, nicklist = ns)

	connect_tab_callbacks(tab, ("joined","topic"))

	return tab


@types(server = TekkaServer, name = basestring)
def create_query(server, name):
	tab = _create_tab(TekkaQuery, name, server)
	return tab


@types (server = basestring)
def create_server(server):
	tab = _create_tab(TekkaServer, server)

	tab.update()

	connect_tab_callbacks(tab, ("new_nick",))

	return tab


@types(server = basestring, child = basestring)
def search_tab(server, child = ""):
	store = widgets.get_object("tab_store")

	for row in store:
		if row[0].name.lower() == server.lower():
			if child:
				for crow in row.iterchildren():
					if crow[0].name.lower() == child.lower():
						return crow[0]
			else:
				return row[0]

	return None


@types (server = basestring, name = basestring)
def search_tabs(server, name=""):
	"""	Searches for a pair of tabs.
		name can be empty, in that case
		only the server string is used
		for the search.

		Possible return values:
		(<serverTab>,<channelTab>)
		(<serverTab>,None)
		(None,None)
	"""
	store = widgets.get_object("tab_store")
	for row in store:
		if row[0].name.lower() == server.lower():
			if not name:
				return (row[0], None)
			else:
				for channel in row.iterchildren():
					if channel[0].name.lower() == name.lower():
						return (row[0], channel[0])
				return (row[0], None)
	return (None, None)


@types (server = (type(None), TekkaServer), object = TekkaTab, update_shortcuts = bool)
def add_tab(server, object, update_shortcuts=True):
	"""	Adds a tab object into the server tree.
		server can be a string identifying a
		server acting as parent for the tab or
		None.

		On succes the method returns the path
		to the new tab, otherwise None.
	"""
	store = widgets.get_object("tab_store")

	serverIter = None

	if server:
		for row in store:
			if row[0] == server:
				serverIter = row.iter

	iter = store.append(serverIter, row=(object,))
	object.path = store.get_path(iter)

	callbacks = get_callbacks("add")

	for cb in callbacks:
		cb(object)

	if server and config.get("tekka", "auto_expand"):
		# expand the whole server tab
		path = store.get_path(serverIter)
		widgets.get_object("tabs_view").expand_row(
			store.get_path(store.iter_parent(iter)), True)

	if update_shortcuts:
		shortcuts.assign_numeric_tab_shortcuts(get_all_tabs())

	return object.path


@types (tab = TekkaTab, update_shortcuts = bool)
def remove_tab(tab, update_shortcuts=True):
	"""	Removes the tab from the server tree.
		There's no need for giving a parent due
		to to the unique identifying path stored
		inner the tekkaTab.
	"""
	store = widgets.get_object("tab_store")

	try:
		row = store[tab.path]
	except IndexError:
		# no tab in server tree at this path
		return False

	# part of hack:
	nextIter = store.iter_next(row.iter)

	if not nextIter:
		temp = store.iter_parent(row.iter)
		if temp:
			nextIter = store.iter_parent(temp)
		else:
			nextIter = None
	path = tab.path

	callbacks = get_callbacks("remove")

	for cb in callbacks:
		cb(tab)

	store.remove(row.iter)
	__updateLowerRows(store, nextIter)

	if update_shortcuts:
		shortcuts.assign_numeric_tab_shortcuts(get_all_tabs())

	return True

def __updateLowerRows(store, iter):
	"""
		iter points to the row after the deleted row.
		path is the path of the deleted row.
		This hack is UGLY! Would someone please fix
		the crappy rows-reordered signal? KTHXBYE
	"""

	if not iter:
		# no work, phew.
		return

	newLastPath = None
	nextIter = iter

	while True:
		if not nextIter:
			break

		tab = store.get(nextIter, 0)
		try:
			tab=tab[0]
		except:
			break

		tab.path = store.get_path(nextIter)

		oIter = nextIter
		nextIter = store.iter_next(oIter)
		if not nextIter:
			if store.iter_has_child(oIter):
				# oIter is a server
				nextIter = store.iter_children(oIter)
			else:
				# oIter is a channel and the next is (maybe)
				# a further server
				temp = store.iter_parent(oIter)
				nextIter = store.iter_next(temp)


@types (old = TekkaTab, new = TekkaTab)
def replace_tab(old, new):
	"""	Replaces the tab `old` with the tab `new`. """
	store = widgets.get_object("tab_store")

	try:
		row = store[old.path]
	except IndexError:
		# no such tab at path
		return False

	store.set(row.iter, 0, new)
	new.path = store.get_path(row.iter)

	# apply new server to childs
	if old.is_server():
		for row in store.iter_children(iter):
			row[0].server = new


def get_tab_by_path(path):
	if not path:
		return

	store = widgets.get_object("tab_store")

	try:
		return store[path][0]
	except (KeyError, IndexError):
		return None


@types (servers = list, excludes = list)
def get_all_tabs(servers=[], excludes=[]):
	"""
		Returns all registered tabs.
		If server is given, only the servers
		and it's children are returned.
		If exclude is given, the given servers
		will be ignored.

		Note:  if there's a newly row inserted, the
		Note:: tab-column can be None.
	"""
	store = widgets.get_object("tab_store")

	tabs = []

	def lower(l):
		return [n.lower() for n in l if n]

	def iterate_store(model, path, iter):
		tab = model[path][0]
		if (None == tab
		or (tab.is_server() and tab.name.lower() in lower(excludes))
		or (
		  not tab.is_server()
		  and tab.server.name.lower() in lower(excludes))):
			return
		tabs.append(tab)


	if not servers:
		store.foreach(iterate_store)
	else:
		for row in store:
			if row[0].name.lower() in lower(servers):
				tabs.append(row[0])
				for child in row.iterchildren():
					tabs.append(child[0])
				break

	return tabs


def get_next_server(current):
	store = widgets.get_object("tab_store")
	useNext = False

	for row in store:
		if useNext:
			return row[0]

		if row[0] == current:
			useNext = True

	if len(store) >= 2:
		# current was the last item, wrap to the first
		return store[0][0]

	return None


@types (tab = TekkaTab)
def get_next_tab(tab):
	""" get the next left tab near to tab.

		This function doesn't consider the
		type of the tab.
	"""
	if not tab or not tab.path:
		return None

	tablist = get_all_tabs()

	if not tablist or len(tablist) == 1:
		return None

	try:
		i = tablist.index(tab)
	except ValueError:
		return None

	return tablist[i-1]


@types(path=tuple)
def switch_to_path(path):
	""" Switch the server tree cursor
		to the row pointed to by path.

		This function returns None.
	"""
	if not mgmt.gui_is_useable:
		return

	tabs_view = widgets.get_object("tabs_view")
	store = widgets.get_object("tab_store")

	if not path:
		logging.error("switch_to_path(): empty path given, aborting.")
		return

	try:
		tab = store[path][0]
	except IndexError:
		logging.error(
			"switch_to_path(): tab not found in store, aborting.")
		return

	old_tab = get_current_tab()

	tabs_view.set_cursor(path)

	# explicit import to not make it accesible to the public
	from current import _set_current_path
	current._set_current_path(path)

	widgets.get_object("output_shell").set(tab.window)

	if tab.is_channel():
		"""
			show up topicbar and nicklist (in case
			they were hidden) and fill them with tab
			specific data.
		"""
		tab.set_useable(tab.joined)

		mgmt.set_user_count(
			len(tab.nickList),
			tab.nickList.get_operator_count())

		mgmt.set_topic(markup.markup_escape(tab.topic))

		if config.get_bool("tekka","show_topic_bar"):

			if (tab.topic
			or not config.get_bool("tekka","hide_topic_if_empty")):
				mgmt.visibility.show_topic_bar(True)
			else:
				mgmt.visibility.show_topic_bar(False)

		mgmt.visibility.show_nicks(True)
		widgets.get_object("nicks_view").set_model(tab.nickList)

	elif tab.is_query() or tab.is_server():
		# queries and server tabs don't have topics or nicklists
		if not status.get("connecting"):
			tab.set_useable(tab.connected)

		if config.get_bool("tekka","show_topic_bar"):
			# hide topic bar in queries if enabled
			mgmt.visibility.show_topic_bar(False)

		# hide nick list in queries
		mgmt.visibility.show_nicks(False)

	# reset message notification
	tab.set_new_message(None)

	mgmt.set_window_title(tab.name)

	if not tab.is_server():
		mgmt.set_nick(tab.server.nick)
	else:
		mgmt.set_nick(tab.nick)

	call_callback("tab_switched", old_tab, tab)


def switch_to_previous():
	tabs = get_all_tabs()
	tab = get_current_tab()

	try:
		i = tabs.index(tab)
	except ValueError:
		return

	try:
		tabs[i-1].switch_to()
	except IndexError:
		return


def switch_to_next():
	tabs = get_all_tabs()
	tab = get_current_tab()

	try:
		i = tabs.index(tab)
	except ValueError:
		return

	try:
		i = i+1
		if (i) == len(tabs):
			i = 0
		tabs[i].switch_to()
	except IndexError:
		return




from tekka.typecheck import types

@types()
def test0(a):
	return True

@types(a = int, b = int)
def test1(a,b):
	return True

@types(a = str, b = str)
def test2(a,b):
	return True

@types(a = (unicode,str), b = (unicode,str))
def test3(a,b):
	return True

if __name__ == "__main__":
	def red(s): return "%s[31m%s%s[0m" % (chr(27),s,chr(27))
	def green(s): return "%s[32m%s%s[0m" % (chr(27),s,chr(27))

	def assert_raise(et, cs):
		try:
			print "Executing '%s'...\t" % (cs),
			eval(cs)

		except BaseException as e:
			if type(e) == et:
				print green("success")
				return

		print red("fail")


	assert( True == test0(1) )

	assert( True == test1(1,2) )
	assert_raise(TypeError, """test1(1.1,2.1)""")
	assert_raise(TypeError, """test1("1","2")""")

	assert( True == test2("a","b") )
	assert_raise(TypeError, """test2(1,2)""")
	assert_raise(TypeError, """test2(u"foo",u"bar")""")

	assert( True == test3(u"foo",u"bar") )
	assert( True == test3("foo", "bar") )
	assert_raise(TypeError, """test3(1,2)""")
	assert_raise(TypeError, """test3(1.0,2.0)""")



# coding: UTF-8
"""
Copyright (c) 2009 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import os
import dbus
import urwid

from gettext import gettext as _
from urwid import MetaSignals
from types import NoneType

from typecheck import types


required_version = (1, 1, 0)


class SushiWrapper(object):

	__metaclass__ = MetaSignals
	signals = ["error","connected","disconnected"]

	@types (sushi_interface = (dbus.Interface, NoneType))
	def __init__(self, sushi_interface):
		self._set_interface(sushi_interface)

	@types (connected = bool)
	def _set_connected(self, connected):
		self._connected = connected

		if connected:
			self._urwid_emit("connected", self)
		else:
			self._urwid_emit("disconnected", self)

	@types (interface = (dbus.Interface, NoneType))
	def _set_interface(self, interface):
		self._sushi = interface
		self._set_connected(interface != None)

	def _urwid_emit(self, signal, *args, **kwargs):
		urwid.emit_signal(self, signal, *args, **kwargs)

	def __getattr__(self, attr):
		def dummy(*args, **kwargs):
			self._urwid_emit("error", _("tekka could not contact maki."),
				_("There's no connection to maki, so the recent "
				"action was not performed. Try to reconnect to "
				"maki to solve this problem."))

		if attr[0] == "_" or attr == "connected":
			# return my attributes
			return object.__getattr__(self, attr)
		else:
			if not self._sushi:
				return dummy
			else:
				if attr in dir(self._sushi):
					# return local from Interface
					return eval("self._sushi.%s" % attr)
				else:
					# return dbus proxy method
					return self._sushi.__getattr__(attr)
		raise AttributeError(attr)

	connected = property(lambda s: s._connected, _set_connected)


sushi = SushiWrapper(None)

_shutdown_handler = None


def get_bus():
	bus_address = os.getenv("SUSHI_REMOTE_BUS_ADDRESS")

	from dbus.mainloop.glib import DBusGMainLoop
	loop=DBusGMainLoop(set_as_default=True)

	if bus_address:
		bus = dbus.connection.Connection(bus_address,mainloop=loop)
	else:
		bus = dbus.SessionBus(mainloop=loop)
	return bus


def connect():
	"""
		Connect to maki over DBus.
		Returns True if the connection attempt
		was succesful.
	"""
	global sushi, _shutdown_handler

	bus = get_bus()

	proxy = None

	try:
		proxy = bus.get_object("de.ikkoku.sushi", "/de/ikkoku/sushi")
	except dbus.exceptions.DBusException, e:
		sushi._urwid_emit(
			"error",
			_("Error while connecting to maki."),
			_("Error: %s" % (e)))

	if not proxy:
		return False

	sushi._set_interface(dbus.Interface(proxy, "de.ikkoku.sushi"))

	version = tuple([int(v) for v in sushi.version()])

	if not version or version < required_version:
		sushi._set_interface(None)
		return False

	_shutdown_handler = sushi.connect_to_signal (
		"shutdown", lambda time: disconnect())

	return True


def disconnect():
	global sushi, __connected, _shutdown_handler

	sushi._set_interface(None)

	_shutdown_handler.remove()


def parse_from (from_str):
	h = from_str.split("!", 2)

	if len(h) < 2:
		return (h[0],"","")

	t = h[1].split("@", 2)

	if len(t) < 2:
		return (h[0],"","")

	return (h[0], t[0], t[1])



# coding: UTF-8
"""
Copyright (c) 2008 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

from .typecheck import types
from .com import sushi

signals = {}
restore_list = []

def setup():
	sushi.g_connect("maki-connected", handle_maki_connect_cb)
	sushi.g_connect("maki-disconnected", handle_maki_disconnect_cb)

"""
One should use this methods to connect to maki's
signals. This API features automatic reconnection
of the registered signals if the connection to
maki was reset.
- connect_signal(<name>,<handler>)
- disconnect_signal(<name>,<handler>)
"""

@types (signal=basestring)
def connect_signal (signal, handler, *userdata):
	""" connect handler to signal """
	global signals

	if not signals.has_key (signal):
	  	signals[signal] = {}

	if signals[signal].has_key(handler):
		# no doubles
		return

	def handler_wrapper(*args, **kwargs):
		handler(*(args+userdata), **kwargs)

	signals[signal][handler] = sushi.connect_to_signal (signal,
														handler_wrapper)


@types (signal=basestring)
def disconnect_signal (signal, handler):
	""" disconnect handler from signal """
	global signals

	try:
		ob = signals[signal][handler]
	except KeyError:
		return
	else:
		ob.remove()
		del signals[signal][handler]


def _restore_signals():
	global restore_list

	for (signal, handler) in restore_list:
		connect_signal(signal, handler)


def handle_maki_disconnect_cb(sushi):
	global signals
	global restore_list

	for signal in signals:
		for handler in signals[signal]:
			h = signals[signal][handler]

			if h:
				h.remove()
				restore_list.append((signal, handler))

	signals = {}


def handle_maki_connect_cb(sushi):
	if restore_list:
		_restore_signals()





# coding: UTF-8
"""
Copyright (c) 2008 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import os
import sys
import logging

# lists are parsed via json
import json

from xdg.BaseDirectory import xdg_config_home, xdg_data_home, \
	xdg_cache_home
import ConfigParser

from .lib import contrast

from .typecheck import types
from .helper.escape import unescape_split, escape_join

from . import update_handler

prefix = ""
defaults = {}

config_parser = None
config_file = ""

encoder = json.JSONEncoder()
decoder = json.JSONDecoder()

_watcher = {}

update_handler = [update_handler.json_get_list_update]

def get_path(*c):
	return os.path.abspath(os.path.join(prefix, *c))

def set_defaults():
	"""
		Sets the default values.

		I know that the ConfigParser class has a
		method to set defaults but these defaults
		are global and not bound to a section.
		If you want a default nick color you can
		set it but if you have another option
		with the same name the default value will
		be the same and this sucks.

		A further point is that this way realizes
		"private" values which are not written
		to config files but can be used with the same
		API. For example the section "uifiles".
	"""
	global defaults

	defaults = {}

	defaults["tekka"] = {}
	defaults["tekka"]["logfile"] = os.path.join(xdg_cache_home, "sushi",
		"tekka.txt")
	defaults["tekka"]["locale_dir"] = get_path("..", "locale")
	defaults["tekka"]["plugin_dirs"] = '["' + '","'.join((
			os.path.join(xdg_data_home, "tekka", "plugins"),
			os.path.join(xdg_data_home, "chirashi"),
			get_path("plugins"),
			get_path("..", "chirashi")
		)) + '"]'
	defaults["tekka"]["use_default_font"] = "True"
	defaults["tekka"]["font"] = "Monospace 10"
        defaults["tekka"]["line_spacing"] = 3
	defaults["tekka"]["auto_expand"] = "True"
	defaults["tekka"]["smooth_scrolling"] = "False"
	defaults["tekka"]["idialog_timeout"] = "False"
	defaults["tekka"]["idialog_timeout_seconds"] = "5"

	defaults["tekka"]["show_general_output"] = "True"
	defaults["tekka"]["show_topic_bar"] = "False"
	defaults["tekka"]["hide_topic_if_empty"] = "True"
	defaults["tekka"]["show_side_pane"] = "True"
	defaults["tekka"]["show_status_bar"] = "True"

	defaults["tekka"]["rgba"] = "False"
	defaults["tekka"]["close_maki_on_close"] = "False"
	defaults["tekka"]["color_text"] = "True" 		# colors enabled?
	defaults["tekka"]["color_nick_text"] = "False" 	# color the text as well as the nick
	defaults["tekka"]["ask_for_key_on_cannotjoin"] = "True"
	defaults["tekka"]["switch_to_channel_after_join"] = "True"

	defaults["tekka"]["whois_dialog"] = "True"      # pop up a dialog on /whois?
	defaults["tekka"]["divider_length"] = "30"      # how much spaces one divider has
	defaults["tekka"]["max_output_lines"] = "500"   # max. lines in one output window
	defaults["tekka"]["profiling"] = "False"        # performance profiling switch
	defaults["tekka"]["popup_line_limit"] = "10"
	defaults["tekka"]["names_columns"] = "5"        # nicks in one row in /names output
	defaults["tekka"]["browser"] = None
	defaults["tekka"]["text_rules"] = "True"
	defaults["tekka"]["rules_limit"] = "3"


	defaults["dcc"] = {}
	defaults["dcc"]["show_ident_in_dialog"] = "False"

	defaults["general_output"] = {}
	defaults["general_output"]["valid_types"] = '["message","action","highlightmessage","highlightaction"]'
	defaults["general_output"]["filter"] = ""
	"""
	filter take place here.
	'type == "message" and server == "euIRC" and channel == "#bsdunix"'
	'not (type == "action" and server == "Freenode" and channel == "#sushi-irc")'
	"""

	defaults["sizes"] = {}
	defaults["sizes"]["outputVPaned"] = "150"
	"""
	[window_height]
	[window_width]
	[$(paned)] = <paned position>
	"""

	defaults["colors"]={}
	defaults["colors"]["own_nick"] = str(contrast.CONTRAST_COLOR_GREY)
	defaults["colors"]["own_text"] = str(contrast.CONTRAST_COLOR_GREY)
	defaults["colors"]["notification"] = str(contrast.CONTRAST_COLOR_LIGHT_GREY)
	defaults["colors"]["text_message"] = str(contrast.CONTRAST_COLOR_BLACK)
	defaults["colors"]["text_action"] = str(contrast.CONTRAST_COLOR_BLACK)
	defaults["colors"]["text_highlightmessage"] = str(contrast.CONTRAST_COLOR_RED)
	defaults["colors"]["text_highlightaction"] = str(contrast.CONTRAST_COLOR_RED)
	defaults["colors"]["nick"] = str(contrast.CONTRAST_COLOR_GREEN) # default foreign nick color
	defaults["colors"]["last_log"] = str(contrast.CONTRAST_COLOR_LIGHT_GREY)

	defaults["colors"]["nick_contrast_colors"] = "True"
	defaults["colors"]["nick_colors"] = '["#AA0000","#2222AA","#44AA44","#123456","#987654"]'
	defaults["colors"]["rules_color"] = "auto" # color for htmlbuffer ruling
	defaults["colors"]["irc_colors"] = "True" # display IRC colors (\003X,Y)
	# markup colors
	defaults["colors"]["new_message"] = str(contrast.CONTRAST_COLOR_LIGHT_BROWN)
	defaults["colors"]["new_action"] = str(contrast.CONTRAST_COLOR_LIGHT_BROWN)
	defaults["colors"]["new_highlightmessage"] = str(contrast.CONTRAST_COLOR_DARK_RED)
	defaults["colors"]["new_highlightaction"] = str(contrast.CONTRAST_COLOR_DARK_GREEN)
	defaults["colors"]["color_new_messages"] = "False"

	defaults["chatting"]={}
	defaults["chatting"]["last_log_lines"] = "10"
	defaults["chatting"]["quit_message"] = "Leading."
	defaults["chatting"]["part_message"] = "Partitioning."
	defaults["chatting"]["nick_separator"] = ": "
	defaults["chatting"]["time_format"] = "%H:%M"
	defaults["chatting"]["highlight_words"] = "[]"

	defaults["autoload_plugins"] = {}

	# section for update handler specific options
	defaults["updates"]={}

	# Add default sections to config parser
	# so setting is easier
	for section in defaults.keys():
		try:
			config_parser.add_section(section)
		except ConfigParser.DuplicateSectionError:
			continue

	# sections defined below are not added to the configParser and
	# can't be set by the set method (will raise NoSectionError)
	defaults["uifiles"] = {}
	defaults["uifiles"]["mainwindow"] = get_path("ui", "main_window.ui")
	defaults["uifiles"]["menus"] = get_path("ui", "menus") + os.path.sep
	defaults["uifiles"]["dialogs"] = get_path("ui", "dialogs") + os.path.sep



def read_config_file():
	"""
	Reads the config file.
	"""
	success = config_parser.read([config_file])

	if not success:
		print "Failed to parse config file '%s'" % config_file
		return False

	return True

def write_config_file():
	"""
	Writes the config values from the
	ConfigParser object into the given file (config_file)
	"""
	if not config_parser:
		logging.error("Config module not loaded. I don't save anything.")
		return

	# remove empty sections so the won't be written
	for section in config_parser.sections():
		if len(config_parser.items(section)) == 0:
			config_parser.remove_section(section)

	f = file(config_file, "w")
	config_parser.write(f)
	f.close()

@types(section=basestring, user_data=tuple)
def add_watcher(section, value, callback, user_data=()):
	""" add watcher to listen for changes on a variable """
	if _watcher.has_key(section):
		if _watcher.has_key(value):
			_watcher[section][value][callback] = user_data
		else:
			_watcher[section][value] = {callback:user_data}
	else:
		_watcher[section] = {value:{callback:user_data}}

def remove_watcher(section, value, callback):
	""" remove a watcher (callback) for this section/value """
	try:
		l = _watcher[section][value]
		i = l.index(callback)
		del l[i]
		return True
	except ValueError, KeyError:
		return False

def _call_watcher(section, value):
	""" call registered watcher for section and value """
	try:
		w = _watcher[section][value]
	except KeyError:
		return
	for (callback,args) in w:
		watcher(section, value, *args)

@types (section = basestring)
def has_section(section):
	return config_parser.has_section(section)

@types (section=basestring)
def create_section(section):
	"""
		creates config section `section`.
	"""
	if not config_parser or config_parser.has_section(section):
		return False
	config_parser.add_section(section)
	return True

@types (section=basestring)
def remove_section(section):
	"""
		removes the section
	"""
	if not config_parser or not config_parser.has_section(section):
		return False

	config_parser.remove_section(section)
	return True

@types (section=basestring, option=basestring)
#@on_fail (print_debug, "ConfigError while setting %s:%s to %s")
def set(section, option, value):
	"""
		Sets in the section the option to value.
		On success the method returns True.
	"""
	if not config_parser:
		return False

	# TODO: enable, test, stuff
	#_call_watcher(section, option)

	try:
		config_parser.set(section, option, str(value))
	except ConfigParser.NoSectionError:
		return False
	else:
		return True

@types(section=basestring, option=basestring, value=bool)
def set_bool(section, option, value):
	return set(section, option, str(value))

@types (section=basestring, option=basestring, l=list)
def set_list(section, option, l):
	""" Set the content of l as value of option.
		Lists are saved in a JSON parseable format.
	"""
	s = encoder.encode(l)

	return set(section, option, s)

@types (section=basestring, option=basestring, value=basestring)
def append_list(section, option, value):
	""" Add value to the list identified by option """
	v = get_list(section, option, [])
	v.append(value)
	return set_list(section, option, v)

@types (section=basestring, option=basestring)
def unset(section, option):
	""" Removes the option in the section.
		Returns True on success otherwise False.
	"""
	if not config_parser:
		return False
	try:
		config_parser.remove_option(section, option)
	except ConfigParser.NoSectionError,e:
		print "Unsetting ('%s','%s') failed: %s" % (section,option,e)
		return False
	return True

@types (section=basestring, option=basestring)
def get(section, option="", default=None):
	"""
		Returns the value for option in section, on
		error the method returns default.

		If option is not given, the whole section
		is returned as dictionary of type {option:value}.
		If there are default values for the section they
		will be merged in.

		The `option`-value is handled case-insensitive.

		get("tekka") will return {"server_shortcuts":"1"}
		get("tekka","server_shortcuts") will return "1"
		get("tekki") will return default
	"""

	if not config_parser:
		return default

	if not option:
		# get the whole section

		if config_parser.has_section(section):
			# the section is noticed by the parser,
			# merge with defaults (if any) and return
			# the value-dict

			new = dict(config_parser.items(section))

			if defaults.has_key(section):
				# merge defaults + config values together
				copy = defaults[section].copy()
				copy.update(new)
				return copy
			else:
				return new

		elif defaults.has_key(section):
			# the config parser does not know the
			# section but it's found in defaults-dict.
			# This is probably a private section.
			return defaults[section]

		else:
			return default

	else:
		# get specific option

		try:
			return config_parser.get(section, option)
		except (ConfigParser.NoOptionError, ConfigParser.NoSectionError),e:
			if defaults.has_key(section):
				try:
					return defaults[section][option]
				except KeyError:
					return default
		else:
			return default

	# usually this is not reached
	return default

@types (section=basestring, option=basestring)
def get_list(section, option, default):
	"""
		Splits the option in the section for ","
		and returns a list if the splitting was
		successful. Else the method will return
		"default".

		Notice: If there's an empty string set,
		this function will return default.
	"""
	res = get(section, option, default)

	if res == default or not res:
		return default

	l = decoder.decode(res)

	if l == default or not l:
		return default
	return l

@types (section=basestring, option=basestring)
def get_bool(section, option, default=False):
	"""
		Returns True or False if the value is
		set or unset.
	"""
	res = get(section, option, default)

	if res == default:
		return default

	if res.lower() == "true" or res == "1":
		return True

	return default

@types (section=basestring, option=basestring)
def get_default(section, option=""):
	"""
	Returns the default value for the option
	in the given section. If no option is given
	(option = None) all defaults of the given
	section are returned as a dictionary.
	If there are no defaults, None is returned.
	"""
	if not option:
		if defaults.has_key(section):
			return defaults[section]
	else:
		if defaults.has_key(section):
			if defaults[section].has_key(option):
				return str(defaults[section][option])
	return None

def is_default(section, option):
	""" check if the value of the option matches the default value
		for the option in the section.
	"""
	return get_default(section,option) == get(section,option)

def reset_value(section, option):
	set(section, option, get_default(section, option))

@types (path=basestring)
def check_config_file(path):
	""" check if config file exists and create it if not """
	if not os.path.exists (path):
		# create the directories
		try:
			os.makedirs (os.path.join (os.path.split (path)[0]))
		except os.error:
			print "Error while creating neccessary directories: %s"\
				% (os.error)
			return False

		try:
			f = file (path, "w")
		except BaseException as e:
			print "Error while creating config file: %s" % (e)
			return False
		else:
			f.close()

		return True
	else:
		return True
	return False

def setup():
	"""
	Find the usual location of the config dir
	(XDG_CONFIG_HOME/sushi/tekka) and parse the
	config file if found.
	"""
	global config_parser, config_file
	global prefix


	if os.path.islink(sys.argv[0]):
		link = os.readlink(sys.argv[0])

		if not os.path.isabs(link):
			link = os.path.join(os.path.dirname(sys.argv[0]), link)

		prefix = os.path.dirname(os.path.abspath(link))
	else:
		prefix = os.path.dirname(os.path.abspath(sys.argv[0]))

	config_parser = ConfigParser.ConfigParser()
	set_defaults()

	config_file = os.path.join (xdg_config_home, "sushi", "tekka")

	if not check_config_file(config_file):
		print "Config file creation failed. Aborting."
		return

	read_config_file()

	for handler in update_handler:
		handler()





from .._builder import widgets

from ...helper import color

"""
Provides utility functions used by (nearly) all tabs.
"""

def _write_to_general_output(msgtype, timestring, tab, message):
	""" write the given data well-formatted to the general output.
		channel can be empty
	"""
	goBuffer = widgets.get_object("general_output").get_buffer()

	if (tab.is_channel() or tab.is_query()):
		# channel print
		goBuffer.go_insert(
						goBuffer.get_end_iter(),
						"[%s] &lt;%s:%s&gt; %s" % (
						   timestring,
						   tab.server.name,
						   tab.name,
						   message),
						tab,
						msgtype)
	else:
		# server print
		goBuffer.go_insert(
						goBuffer.get_end_iter(),
						"[%s] &lt;%s&gt; %s" % (
						   timestring,
						   tab.name,
						   message),
						tab,
						msgtype)

	widgets.get_object("general_output").scroll_to_bottom()


def _markup_color(key):
	return color.get_color_by_key(key,
		color.get_widget_base_color(widgets.get_object("tabs_view")))



"""
Copyright (c) 2009-2010 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gtk
import os

from ._builder import widgets

from .. import config
from ..typecheck import types

from ..lib.htmlbuffer import HTMLBuffer
from ..lib.status_icon import TekkaStatusIcon
from ..lib.output_window import OutputWindow
from ..lib.output_shell import OutputShell
from ..lib.search_toolbar import SearchBar
from ..lib.expanding_list import ExpandingList
from ..lib.contrast_color_table import ContrastColorTable
from ..lib.custom_color_button import CustomColorButton

from ..helper import URLHandler

"""
Creating/Bootstrapping module. This is the core of the UI creation.

All gtk.Builder files are handled here.

Therefore, gtk.Builder needs to find all dependend UI modules
so they are imported here.
"""


def get_new_buffer():
	""" Returns a HTMLBuffer with assigned URL handler. """
	buffer = HTMLBuffer(handler=URLHandler.URLHandler)
	return buffer


def get_new_output_window():
	w = OutputWindow()
	return w


def information_dialog(title="", message=""):
	""" create a dialog with an info icon, a title and a message.
		This dialog has one button (OK) and does not handle it's response.
	"""
	d = gtk.MessageDialog(
		   		  type = gtk.MESSAGE_INFO,
			   buttons = gtk.BUTTONS_OK,
		message_format = message)

	d.set_title(title)

	return d


def question_dialog(title = "", message = ""):
	""" create a dialog with a question mark, a title and a message.
		This dialog has two buttons (yes, no) and does not handle
		it's response.
	"""
	d = gtk.MessageDialog(
		   		  type = gtk.MESSAGE_QUESTION,
			   buttons = gtk.BUTTONS_YES_NO,
		message_format = message)

	d.set_title(title)

	return d


def error_dialog(title = "", message = ""):
	""" create a dialog with a exclamation mark, a title and a message.
		This dialog has one close button and does not handle it's
		response.
	"""
	err = gtk.MessageDialog(
				  type = gtk.MESSAGE_ERROR,
			   buttons = gtk.BUTTONS_CLOSE,
		message_format = message)

	err.set_title(title)

	return err


def build_status_icon():
	"""
	Sets up the status icon.
	"""
	if config.get_bool("tekka", "rgba"):
		gtk.widget_push_colormap(
			widgets.get_object("main_window")\
			.get_screen().get_rgb_colormap()
		)

	statusIcon = TekkaStatusIcon()
	widgets.add_object(statusIcon, "status_icon")

	if config.get_bool("tekka", "rgba"):
		gtk.widget_pop_colormap()

	statusIcon.set_visible(True)


@types(ui_file=basestring, section=basestring)
def load_main_window(ui_file):
	""" Load the widgets from the UI file, do simple setup on them,
		setup the widgets wrapper and return it.
		After succesful setup, load-finished is emitted.
	"""

	widgets.add_from_file(ui_file)

	def setup_mainmenu_context():
		from ..menus.mainmenu_context import MainMenuContext
		return MainMenuContext(name="menubar", widgets=widgets)

	mainmenu = setup_mainmenu_context()
	widgets.add_object(mainmenu, "main_menu_context")

	return widgets


def load_menu(name):
	# menus are gtkbuilder
	path = os.path.join(
					config.get("uifiles", "menus"),
					name + ".ui")

	builder = gtk.Builder()

	builder.add_from_file(path)

	return builder


def load_dialog(name):
	path = os.path.join(
		config.get("uifiles", "dialogs"),
		name + ".ui"
	)

	builder = gtk.Builder()
	builder.add_from_file(path)

	return builder


class Builder(gtk.Builder):

	def load_dialog(self, name):
		path = os.path.join(config.get("uifiles", "dialogs"),
							name + ".ui")
		self.add_from_file(path)

	def load_menu(self, name):
		path = os.path.join(config.get("uifiles", "menus"),
							name + ".ui")
		self.add_from_file(path)




# coding: UTF-8
"""
Copyright (c) 2008 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import commands
import config
import tabs

# types identificating the current scroll position
(
	NO_TYPE,
	NICK_TYPE,
	QUERY_TYPE,
	CHANNEL_TYPE,
	COMMAND_TYPE
) = range(5)

# cache of position
_current = {
	"position": None,
	"tab": None,
	"type": NO_TYPE,
	"needle": None,
	"lastCompletion": None
}

# reference to the main window
_main_window = None

def _reset_iteration():
	global _current
	_current["position"] = None
	_current["tab"] = None
	_current["type"] = NO_TYPE
	_current["needle"] = None
	_current["lastCompletion"] = None

def _appendMatch(entry, text, word, match, separator=" "):
	""" Complete `word` in `text` with `match` and
		apply it to the input bar widget.
		Add separator too.
	"""
	# old text without needle + new_word + separator + rest
	new_text = text[0:entry.edit_pos-len(word)] + \
		match + \
		separator + \
		text[entry.edit_pos:]

	old_position = entry.edit_pos
	entry.set_edit_text(new_text)

	entry.set_edit_pos(old_position + len(match+separator) -len(word))

	global _current
	_current["lastCompletion"] = match + separator

def _removeLastCompletion(entry, text):
	""" this function assumes, we're on the position _after_
		the completion...
	"""
	lc = _current["lastCompletion"]

	if lc == None:
		return text

	# strip of the match, keep the needle:
	# 'n'<Tab> => 'nemo: ' => strip 'emo: '
	needle = _current["needle"]
	skip = (entry.edit_pos - len(lc)) + len(needle)
	new_text = text[:skip]+text[entry.edit_pos:]

	entry.set_edit_text(new_text)
	entry.set_edit_pos(skip)

	return new_text

def _raise_position(matches, i_type):
	if _current["type"] == i_type:
		# continue iterating
		if (_current["position"]+1 >= len(matches)
			or _current["position"] == None):
			_current["position"] = 0

		else:
			_current["position"] += 1

	else:
		# set type to the current and begin iterating
		_current["type"] = i_type
		_current["position"] = 0

def _match_nick_in_channel(tab, word):
	matches = [n for n in tab.nicklist.keys() if n[:len(word)].lower() == word.lower()]
	# sort nicks alphabetically
	matches.sort(lambda x, y: cmp(x.lower(), y.lower()))

	if matches:
		_raise_position(matches, NICK_TYPE)
		return matches[_current["position"]]
	return None

def _match_nick_in_query(tab, word):
	matches = [nick for nick in (tab.name, (tab.is_server() and tab.nick or tab.parent.nick)) if nick[:len(word)].lower() == word.lower()]

	if matches:
		_raise_position(matches, QUERY_TYPE)
		return matches[_current["position"]]
	return None

def _match_channel(word):
	global _main_window

	tablist = tabs.tree_to_list(_main_window.servers)

	# find all matching tabs
	matches = [tab.name for tab in tablist
		if tab and tab.name[:len(word)].lower() == word.lower()]

	if matches:
		_raise_position(matches, CHANNEL_TYPE)
		return matches[_current["position"]]
	return None

def _match_command(word):
	matches = [cmd for cmd in commands._command_dict.keys()
	if cmd[:len(word)].lower()==word.lower()]

	if matches:
		_raise_position(matches, COMMAND_TYPE)
		return matches[_current["position"]]
	return None

def stopIteration():
	""" user does not want any more results, stop the iteration.
		This is the case if, for example, the tab is switched or
		the input bar is activated.
	"""
	_reset_iteration()

def setup(main_window):
	global _main_window
	_main_window = main_window

def complete(entry, text):
	""" search for the last typed word and try to
		complete it in the following order:
		- search for a suitable nick in the channel (if tab is a channel)
		- search for a suitable nick in the query (if tab is a query)
		- search for a suitable command (if the first letter is a '/')
		- search for a suitable channel (if the first letter is
		  a valid channel prefix)

		If one of the searches matches, this function returns True.
		If no search matches, False is returned.

		The function checks, if the word searched for was the
		same as last time. So if complete is called another
		time, it will continue searching and using the next
		result to the one before.
	"""

	global _current

	currentTab = _main_window.current_tab

	if not text:
		return False

	if currentTab != _current["tab"]:
		# reset iteration, data is not up to date.
		# start new iteration, then
		_reset_iteration()
		_current["tab"] = currentTab

	if _current["needle"]:
		# continue searching for `needle`
		word = _current["needle"]
		text = _removeLastCompletion(entry, text)

	else:
		# get the word to complete
		# "f|<tab> || f |<tab>"
		word = text[0:entry.edit_pos].split(" ")[-1].strip()

		_current["needle"] = word

	if not word:
		return False

	match = None

	if currentTab and type(currentTab) == tabs.Channel:
		# look for nicks

		match = _match_nick_in_channel(currentTab, word)

		if match:

			# if the word is in a sentence, use command
			# completion (only whitespace)
			if text.count(" ") < 1:
				separator = config.get("chatting", "nick_separator")
			else:
				separator = " "

			_appendMatch(entry, text, word, match,
				separator = separator)

			return True

	elif currentTab and type(currentTab) == tabs.Query:
		# look for my nick or the other nick

		match = _match_nick_in_query(currentTab, word)

		if match:
			if text.count(" ") < 1:
				separator = config.get("chatting", "nick_separator")
			else:
				separator = " "

			_appendMatch(entry, text, word, match,
				separator = separator)

			return True

	# ** no successful completion so far **

	# channel completion
	if (currentTab
	and word[0] in (type(currentTab) == tabs.Server
					and currentTab.support_chantypes
					or currentTab.parent.support_chantypes)):

		match = _match_channel(word)

		if match:
			_appendMatch(entry, text, word, match)
			return True

	# *** command completion ***

	if word[0] == "/":

		match = _match_command(word[1:])

		if match:
			_appendMatch(entry, text, word, "/"+match)
			return True

	return False



import gtk
import gobject

from dbus import UInt64

from ... import config
from ... import com # lastlog

from ...helper import color
from ...helper import markup
from ...typecheck import types

from .messages import *
from .current import *

"""
Provides the basic tab with the basic functionality.

All other tabs (server, channel, query) are derived from it.
"""

class TekkaTab(gobject.GObject):
	"""
		Provides basic attributes like the output textview,
		the name of the tab and a flag if a new message is received.

		Attributes:
		textview: the textview bound to this tag
		path: the identifying path in gtk.TreeModel
		name: the identifying name
		newMessage: a list containing message "flags"
		connected: is the tab active or not
	"""

	@types(switch=bool)
	def _set_connected(self, switch):
		self._connected=switch
		self.emit ("server_connected", switch)
		self.emit ("new_markup")
	connected = property(lambda x: x._connected, _set_connected)


	@types(path=tuple)
	def _set_path(self, path):
		self._path = path
	path = property(lambda x: x._path, _set_path)


	@types(name=basestring)
	def _set_name(self, name):
		self._name = name
		self.emit ("new_name", name)
	name = property(lambda x: x._name, _set_name)


	def __init__(self, name, window = None):
		gobject.GObject.__init__(self)

		self.window = window   # the associated GUI output widget
		self.path = ()         # the path in the server tree
		self.name = name       # identifying name
		self.newMessage = []   # status array of unread message types
		self.connected = False # status flag if the tab's server is connected
		self.input_text = ""   # last typed input text

		self.input_history = None # input_history object placeholder


	def __repr__(self):
		return "<tab '%s', path: '%s'>" % (self.name, self.path)


	def is_server(self):
		return False


	def is_query(self):
		return False


	def is_channel(self):
		return False


	@types(text = str)
	def set_input_text(self, text):
		self.input_text = text


	def get_input_text(self):
		return self.input_text


	@types(status = (str, type(None)))
	def set_new_message(self, status):
		""" A message is unique set and represents
			the status of the tab. The message stack
			can be reset by using None as status.

			See MSGTYPES at the beginning of the file
			for a listing of available message types.
		"""
		new = False

		if not status:
			self.newMessage = []
		else:
			try:
				self.newMessage.index(status)
			except ValueError:
				self.newMessage.append(status)
				new = True
			self.emit ("new_message", status)

		if not status or new:
			self.emit("new_markup")


	def markup(self):
		if self.newMessage:
			return "<b>"+self.name+"</b>"
		return self.name


	def write(self, *x):
		raise NotImplementedError


	def write_raw(self, msg, type=MESSAGE):
		""" unformatted, without timestamp """
		buf = self.window.textview.get_buffer()
		end = buf.get_end_iter()

		buf.insert_html(end, msg)

		def notify():
			self.set_new_message(type)
			return False

		gobject.idle_add(notify)


	def print_last_log(self, lines=0):
		"""	Fetch the given amount of lines of history for
			the channel on the given server and print it to the
			channel's textview.
		"""
		from .server import TekkaServer

		# XXX: instead of catching this, cancel new()
		if type(self) == TekkaTab:
			raise NotImplementedError

		buffer = self.window.textview.get_buffer()

		lines = UInt64(lines or config.get(
			"chatting",
			"last_log_lines"))

		if type(self) == TekkaServer:
			server = self.name
			channel = ""
		else:
			server = self.server.name
			channel = self.name

		for line in com.sushi.log(server, channel, lines):
			line = color.strip_color_codes(line)

			buffer.insert_html(buffer.get_end_iter(),
				"<font foreground='%s'>%s</font>" % (
					color.get_color_by_key("last_log"),
					markup.escape(line)))


	def set_useable(self, switch):
		widgetList = [
			widgets.get_object('nicks_view'),
			widgets.get_object("output")]

		for widget in widgetList:
			if widget.get_property("sensitive") == switch:
				continue
			widget.set_sensitive (switch)


	def is_active(self):
		serverTab, channelTab = get_current_tabs()

		if not serverTab or (not channelTab and self.is_channel()):
			return False

		if (self.is_server()
			and self.name.lower() == serverTab.name.lower()
			and not channelTab):
			return True

		if (( self.is_channel() or self.is_query() )
			and channelTab
			and self.name.lower() == channelTab.name.lower()
			and self.server.name.lower() == serverTab.name.lower()):
			return True

		return False


	def switch_to(self):
		" convenience method for tabs.switch_to_path "
		from ..tabs import switch_to_path
		switch_to_path(self.path)


gobject.signal_new(
	"server_connected",
	TekkaTab,
	gobject.SIGNAL_ACTION,
	gobject.TYPE_NONE,
	(gobject.TYPE_BOOLEAN,))

""" The tab's markup has changed """
gobject.signal_new(
	"new_markup",
	TekkaTab,
	gobject.SIGNAL_ACTION,
	gobject.TYPE_NONE,())

""" The tab's unread messages buffer changed """
gobject.signal_new(
	"new_message",
	TekkaTab,
	gobject.SIGNAL_ACTION,
	gobject.TYPE_NONE,
	(gobject.TYPE_PYOBJECT,))

""" The tab's name changed (new name as parameter) """
gobject.signal_new(
	"new_name",
	TekkaTab,
	gobject.SIGNAL_ACTION,
	gobject.TYPE_NONE,
	(gobject.TYPE_STRING,))




"""
Handler for updates between revisions and versions.
"""

import gobject
import logging

from . import gui

def message_success(user_data):
	(title, message) = user_data
	gui.mgmt.show_inline_message(title, message)

def message_failure(user_data):
	(title, message) = user_data
	gui.mgmt.show_inline_message(title, message, dtype="error")


def json_get_list_update():
	" get_list internals were replaced by json parsing, updates config "

	from . import config
	from .helper import escape

	if config.get_bool("updates", "json_config"):
		return

	logging.info("json_get_list_update starting.")

	changes = 0

	for section in config.config_parser.sections():
		for (name, value) in config.config_parser.items(section):
			if value.count(",") > 1 and value[0] != "[": # got old list
				# convert to list via old stuff
				l = escape.unescape_split(",", value)
				config.set_list(section, name, l)

				changes += 1

				logging.info("Updated option %s from %s to %s" % (
					name, value, config.get_list(section, name, [])))

	logging.info("json_get_list_update done: %d changes." % (changes,))

	config.set_bool("updates", "json_config", True)

	gobject.idle_add(message_success, ("Update succeeded",
		"The config update was successfuly applied."))



import gobject
import time

from ... import com # parse_from, sushi
from ... import config

from ...typecheck import types

from . import tab
from . import util

from .messages import *
from .current import *

class TekkaServer(tab.TekkaTab):

	@types(msg=basestring)
	def _set_away(self, msg):
		self._away = msg
		self.emit("away", msg)
		self.emit("new_markup")


	@types(prefix = (tuple, list))
	def _set_sprefix(self, prefix):
		self._sprefix = prefix
		self.emit("prefix_set", prefix)


	@types(ctypes = (tuple, list, basestring))
	def _set_ctypes(self, ctypes):
		self._ctypes = ctypes
		self.emit("channeltypes_set", ctypes)


	@types(nick = basestring)
	def _set_nick(self, nick):
		self._nick = nick
		self.emit("new_nick", nick)


	support_prefix = property(lambda x: x._sprefix, _set_sprefix)
	support_chantypes = property(lambda x: x._ctypes, _set_ctypes)
	away = property(lambda x: x._away, _set_away)
	nick = property(lambda x: x._nick, _set_nick)


	def __init__(self, name, textview=None):
		super(TekkaServer,self).__init__(name, textview)

		self.nick = ""               # IRC nick
		self.away = ""               # Away message
		self.support_prefix = ()     # Which prefixed are supported
		self.support_chantypes = ()  # Which chantypes are supported


	def is_server(self):
		return True


	def markup(self):
		base = self.name

		if not self.connected:
			base = "<span strikethrough='true'>"+base+"</span>"

		if self.newMessage:
			base = "<b>"+base+"</b>"

		if self.away:
			base = "<i>"+base+"</i>"

		return base


	def write(self, timestamp, message, msgtype=MESSAGE,
		no_general_output=False):
		""" write [<timestamp>] <message> to this tab's buffer """

		buffer = self.window.textview.get_buffer()

		timestr = time.strftime(
						config.get("chatting", "time_format", "%H:%M"),
						time.localtime(timestamp))

		buffer.insert_html(
			buffer.get_end_iter(),
			"[%s] %s" % (timestr, message))

		if not self.is_active():

			if (config.get_bool("tekka", "show_general_output")
			and not no_general_output):

				util._write_to_general_output(
					msgtype,
					timestr,
					self,
					message)

		def notify():
			self.set_new_message(msgtype)
			return False

		gobject.idle_add(notify)


	def current_write(self, timestamp, message, msgtype=MESSAGE,
	no_general_output=False):
		""" write a string to the current active tab of
			this server or, if no tab is active, to the
			this tab
		"""

		cServerTab, cChannelTab = get_current_tabs()

		if (cServerTab and cServerTab.name.lower() == self.name.lower()
		and cChannelTab):

			cChannelTab.write(
				timestamp,
				message,
				msgtype,
				no_general_output)
		else:

			self.write(
				timestamp,
				message,
				msgtype,
				no_general_output)


	def update(self):
		""" fetch server info from sushi and apply them
			to this tab
		"""
		server = self.name

		if com.sushi.user_away(server, com.get_own_nick(server)):
			# TODO: better query management @ LP
			self.away = "-- Not implemented yet --"

		self.nick = com.parse_from(com.sushi.user_from(server, ""))[0]
		self.support_prefix = com.sushi.support_prefix(server)
		self.support_chantypes = com.sushi.support_chantypes(server)

""" Away status changed (message as parameter) """
gobject.signal_new(
	"away",
	TekkaServer, gobject.SIGNAL_ACTION,
	gobject.TYPE_NONE, (gobject.TYPE_STRING,))

""" Supported chantypes set (tuple as parameter) """
gobject.signal_new(
	"channeltypes_set",
	TekkaServer, gobject.SIGNAL_ACTION,
	gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT,))

""" Supported prefixed set (tuple as parameter) """
gobject.signal_new(
	"prefix_set",
	TekkaServer, gobject.SIGNAL_ACTION,
	gobject.TYPE_NONE, (gobject.TYPE_PYOBJECT,))

""" IRC nick changed (nick as parameter) """
gobject.signal_new(
	"new_nick",
	TekkaServer, gobject.SIGNAL_ACTION,
	gobject.TYPE_NONE, (gobject.TYPE_STRING,))





# coding: UTF-8
"""
Copyright (c) 2008 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gtk
import gobject
import logging


class ExpandingList(gtk.Table):

	__gtype_name__ = "ExpandingList"

	# add ability of setting widgets as tuple represended by a string.
	# This is a hack for glade support
	def set_str_widgets(self, str_widgets):
		try:
			widgets = eval(str_widgets)
		except Exception as e:
			logging.debug("Error in set_str_widgets: %s" % (e,),
						  exc_info=True)
			return
		self.init_widgets(widgets)
	str_widgets = gobject.property(setter=set_str_widgets, type=str)
	
	def set_no_first_row(self, no_first_row):
		self._no_first_row = no_first_row
	no_first_row = gobject.property(setter=set_no_first_row, default=False, type=bool)

	def __init__(self, no_first_row=False, *widgets, **kwargs):

		self._rows = 1
		self._columns = len(widgets) + 2
		self._matrix = [ [] ]
		self._widgets = ()
		self._no_first_row = no_first_row

		gtk.Table.__init__(self, rows=1, columns=self._columns)

		if not widgets:
			return

		init_widgets(widgets)


	def init_widgets(self, widgets):
		self._widgets = widgets

		if not self._no_first_row:
			self._add_row(row=0)

	def get_widget_matrix(self):
		return self._matrix

	def _add_row(self, row):
		"""
		Adds a row filled with the given widgets
		(self._widgets) and a plus (add) as well as
		a minus (remove) button. It also extends
		the matrix which holds the widgets.
		"""


		self._matrix[row]=[]
		column = 0

		for widget in self._widgets:

			try:
				instance = widget()
			except:
				logging.error(
					"ExpandingList: error while instancing %s" % (widget))
				continue
			self.emit("instanced_widget", row, column, instance)

			logging.debug("attaching instance of %s" % (widget))
			self.attach(instance, column, column+1, row, row+1)
			self._matrix[row].append(instance)
			column += 1

		self._add_plus_button(row, column)
		self._add_minus_button(row, column+1)

		self.emit("row-added", row)

		self.show_all()

	def add_row(self, under=-1):
		"""
		under is the index (starting at 0)
		under which row the new one shall be added.
		if under is not given or is under 0,
		the new row is added under the last one.
		"""
		if not self._widgets:
			return

		# determine the row to add the new row under
		if under >= 0:
			if under > self._rows:
				logging.error(
					"expanding_list: add_row: under (%d) > "
					"self._rows (%d)" % (under, self._rows))
				return

			row = under + 1
		else:
			row = self._rows

		self._rows += 1
		self.resize(self._rows, self._columns)
		self._matrix.append([])

		"""
		rows = 5; under = 2
		last row is newly added
		  _____
		0| | | |
		1| | | |
		2| | | |------.
		3| | | |--. <-'
		4|_|_|_|<-'
		"""
		if row < (self._rows-1):
			# swap needed

			for i in reversed(range(row, self._rows-1)):
				cells = self._matrix[i]

				column = 0
				for widget in cells:
					self.remove(widget)
					self.attach(widget, column, column+1, i+1, i+2)
					column += 1

				cells[-1].row = i+1
				cells[-2].row = i+1

				self._matrix[i+1] = self._matrix[i]

		self._add_row(row)

	def remove_row(self, index):
		"""
		Remove the row with the given index from
		the table.
		"""
		if not self._widgets:
			return

		if index > (self._rows - 1) or index < 0:
			raise Exception("index out of bounds")

		if self._rows == 1:
			# replace the last row with a new, empty one
			self.add_row(under=0)
			self.remove_row(0)
			return

		cells = self._matrix[index]

		# remove the widgets in the destinasted row
		for widget in cells:
			self.remove(widget)

		for i in range(index+1, self._rows):
			# i -> (i-1)
			row = self._matrix[i]
			newRow = i - 1

			column = 0
			for widget in row:
				self.remove(widget)
				self.attach(widget, column, column+1, newRow, i)
				column += 1

			row[-1].row = newRow
			row[-2].row = newRow

			self._matrix[newRow] = self._matrix[i]

		# NOW remove the line
		del self._matrix[-1]

		self.emit("row-removed", index)

		# bring the table to the new size
		self._rows -= 1
		self.resize(self._rows, self._columns)

	def attach(self, *x):
		if not x:
			return
		widget = x[0]

		gtk.Table.attach(self, *x)

		if widget.parent:
			self.child_set_property(widget, "y-options", gtk.SHRINK)

	def _button_add_clicked(self, button):
		self.add_row(under=button.row)

	def _button_remove_clicked(self, button):
		self.remove_row(button.row)

	def _add_plus_button(self, row, column):
		a = gtk.Button(stock=gtk.STOCK_ADD)
		# remove the label of the button
		a.get_children()[0].get_children()[0].get_children()[1].set_text("")
		# show icon
		a.get_image().show()

		a.row = row
		a.column = column

		a.connect("clicked", self._button_add_clicked)
		self._matrix[row].append(a)
		self.attach(a, column, column+1, row, row+1, gtk.FILL|gtk.SHRINK, gtk.FILL|gtk.SHRINK)

	def _add_minus_button(self, row, column):
		a = gtk.Button(stock=gtk.STOCK_REMOVE)
		# remove the label of the button
		a.get_children()[0].get_children()[0].get_children()[1].set_text("")
		# show icon
		a.get_image().show()

		a.row = row
		a.column = column

		a.connect("clicked", self._button_remove_clicked)
		self._matrix[row].append(a)
		self.attach(a, column, column+1, row, row+1, gtk.FILL|gtk.SHRINK, gtk.FILL|gtk.SHRINK)

gobject.signal_new("instanced_widget", ExpandingList, gobject.SIGNAL_ACTION,
	gobject.TYPE_NONE, (gobject.TYPE_INT, gobject.TYPE_INT, gobject.TYPE_PYOBJECT))

gobject.signal_new("row-added", ExpandingList, gobject.SIGNAL_ACTION,
	gobject.TYPE_NONE, (gobject.TYPE_INT,))

gobject.signal_new("row-removed", ExpandingList, gobject.SIGNAL_ACTION,
	gobject.TYPE_NONE, (gobject.TYPE_INT,))

if __name__ == "__main__":
	win = gtk.Window()
	win.resize(400,400)

	sWin = gtk.ScrolledWindow()

	expList = ExpandingList(gtk.Entry, gtk.Entry)
	expList.set_property("homogeneous", True)

	sWin.add_with_viewport(expList)

	win.add(sWin)
	win.connect("destroy", lambda *x: gtk.main_quit())

	win.show_all()

	gtk.main()




# coding: UTF-8
"""
Copyright (c) 2008-2010 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gtk
import logging
from gettext import gettext as _

from .. import com
from .. import gui

widgets = None
serverSelection = None

RESPONSE_CONNECT = 3


def setup():
	global widgets, serverSelection

	if widgets:
		return

	widgets = gui.builder.load_dialog("servers")

	sigdic = {  "addButton_clicked_cb":
					lambda w: openAddDialog(),
				"editButton_clicked_cb":
					lambda w: openEditDialog(),
				"deleteButton_clicked_cb":
					lambda w: openDeleteDialog(),
				"serverRenderer_edited_cb":
					serverNameEdited
			 }

	widgets.connect_signals(sigdic)

	# enable multiple selection
	serverSelection = widgets.get_object("serverList").get_selection()
	serverSelection.set_mode(gtk.SELECTION_MULTIPLE)


def addServer(name):
	""" Add server from maki to the list store """
	widgets.get_object("serverStore").append([name])


def retrieveServerlist():
	""" Fetch server list from maki and get
		infos about every server.
	"""
	widgets.get_object("serverStore").clear()

	servers = com.sushi.server_list("","")

	if servers:
		for server in servers:
			addServer(server)


def serverNameEdited(renderer, path, newText):
	""" User edited column in serverView """

	try:
		oldText = widgets.get_object("serverStore")[path][0]
	except IndexError:
		return

	com.sushi.server_rename(oldText, newText)

	# at last, update the list from maki (caching would be better..)
	retrieveServerlist()


def run(callback):
	dialog = widgets.get_object("serverDialog")

	retrieveServerlist()

	main_window = gui.mgmt.widgets.get_object("main_window")
	dialog.set_transient_for(main_window)

	dialog.connect("response", dialog_response_cb, callback)
	dialog.show()


def createServer(serverName, data):
	""" Create a server in maki. """
	for (k,v) in data.items():
		com.sushi.server_set(serverName, "server", k, v)


def deleteServer(servername):
	""" Remove server from Serverlist widget
		and delete server in maki.
	"""
	serverList = widgets.get_object("serverStore")

	for row in serverList:
		if row[0] == servername:
			serverList.remove(row.iter)
			com.sushi.server_remove(servername, "", "")


def openAddDialog():
	gui.dialogs.show_dialog("addServer", add_dialog_cb)


def openEditDialog():
	view = widgets.get_object("serverList")
	serverList = view.get_model()

	path = view.get_cursor()[0]

	servername = None

	if not path:
		d = gui.builder.information_dialog(
								_("No server selected."),
								_("You must select a server to edit it."))
		d.connect("response", lambda w,i: w.destroy())
		d.show_all()
		return

	else:
		servername = serverList[path][0]

	data = gui.dialogs.show_dialog("editServer", servername)

	if not servername:
		logging.error("openEditDialog: Error in retrieving the servername")
		return

	if data:
		retrieveServerlist()


def openDeleteDialog():
	view = widgets.get_object("serverList")

	path = view.get_cursor()[0]
	servername = None

	if not path:
		d = gui.builder.information_dialog(
								_("No server selected."),
								_("You must select a server to delete it."))
		d.connect("response", lambda w,i: w.destroy())
		d.show_all()
		return

	else:
		servername = view.get_model()[path][0]

	if not servername:
		gui.mgmt.show_error_dialog(
			title=_("Error while retrieving server name."),
			message=_("There was an error while retrieving the server "
					  "name.\nAre you connected to maki?"))
		return

	gui.dialogs.show_dialog("deleteServer", servername, delete_dialog_cb)


def dialog_response_cb(dialog, response_id, callback):
	if response_id == RESPONSE_CONNECT:
		# get the selected server(s)

		serverList = widgets.get_object("serverStore")
		paths = serverSelection.get_selected_rows()[1]

		if not paths:
			return

		toConnect = []
		for path in paths:
			toConnect.append(serverList[path][0])

		callback(toConnect)

	else:
		callback(None)

	dialog.hide()


def add_dialog_cb():
	""" indicates, server was added """
	retrieveServerlist()


def delete_dialog_cb(servername):
	""" indicates that the server can be deleted """
	deleteServer(servername)





"""
Copyright (c) 2009-2010 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gobject

"""
Holds the StatusManager, an object which holds a status application-wide.

This is used for the status bar and error handling.
"""

class StatusManager(gobject.GObject):

	""" Hold the different states the application is in.
		Every state has an ID which can be retrieved by
		id(status). A status can be set or be unset or
		retrieved by get(status)
	"""

	def __init__(self):
		gobject.GObject.__init__(self)
		self.states = []

	def set(self, status):
		""" Set the given status """
		try:
			self.states.index(status)
		except ValueError:
			self.states.append(status)
			self.emit("set-status", status)
			return True
		return False

	def set_visible(self, status, message):
		""" Set the given status with a message which can be processed """
		if self.set(status):
			self.emit("set-visible-status", status, message)

	def unset(self, status):
		""" Unset the given status """
		try:
			index = self.states.index(status)
		except ValueError:
			return False
		self.emit("unset-status", status)
		del self.states[index]
		return True

	def get(self, status):
		""" return True if status is set, otherwise False """
		try:
			self.states.index(status)
		except ValueError:
			return False
		else:
			return True

	def id(self, status):
		""" return an unique ID for the given status if it's set.
			Otherwise raise a ValueError
		"""
		try:
			return self.states.index(status)
		except ValueError:
			raise ValueError, "Status %s not in state list." % (status)

gobject.signal_new("set-status", StatusManager, gobject.SIGNAL_ACTION,
	None, (gobject.TYPE_STRING,))
gobject.signal_new("set-visible-status", StatusManager,
	gobject.SIGNAL_ACTION, None, (gobject.TYPE_STRING, gobject.TYPE_STRING))
gobject.signal_new("unset-status", StatusManager, gobject.SIGNAL_ACTION,
	None, (gobject.TYPE_STRING,))

status = StatusManager()



# coding: UTF-8
"""
Copyright (c) 2008 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import os
import re
import gobject

import dbus
from dbus.mainloop.glib import DBusGMainLoop
from gettext import gettext as _
from types import NoneType

from tekka.typecheck import types


"""
The com module caches the nick,
provides a persistent interface to maki,
provides methods to connect / disconnect
maki and to send messages.
"""


class NoSushiError (BaseException):
	pass


class SushiWrapper (gobject.GObject):

	""" Wraps a DBus Interface to maki so if there's
		no connection, an error signal is emitted and
		can be catched by the GUI.

		Access to the underlying gobject methods is
		possible by prefixing an "g_":
		g_emit(), g_connect, ...

		The only methods available directly from
		this class are all methods beginning with an "_"
		and the "connected" property.
		All other getattr calls will be forwarded
		to the intern sushi interface.
	"""


	@types (sushi_interface = (dbus.Interface, NoneType))
	def __init__(self, sushi_interface):
		gobject.GObject.__init__(self)

		self.connected = False
		self.remote = False

		self._set_interface(sushi_interface)


	@types(v = bool)
	def _set_remote(self, v):
		self._remote = v


	@types (connected = bool)
	def _set_connected(self, connected):
		if connected:
			self.g_emit("maki-connected")
		else:
			self.g_emit("maki-disconnected")

		self.__recent_reconnect = True
		self._connected = connected


	@types (interface = (dbus.Interface, NoneType))
	def _set_interface(self, interface):
		self._sushi = interface
		self._set_connected(interface != None)


	@types (title = basestring, msg = basestring)
	def _emit_error(self, title, msg):
		self.g_emit("sushi-error", title, msg)


	def __getattribute__(self, attr):

		""" attributes prefixed with g_ will be resolved
			by GObject's getattribute with the g_ striped.

			attributes prefixed with _ will be resolved
			as they are for SushiWrapper.

			all other attributes are forwarded to self._sushi
		"""

		def dummy(*args, **kwargs):
			""" called if no connection is active """

			if self.__recent_reconnect:
				self.__recent_reconnect = False

				sushi._emit_error(
					_("tekka could not connect to maki."),
					_("Please check whether maki is running."))

			pass

		def errordummy(message):
			""" called if a DBusError was catched """

			def new(*args, **kwargs):
				sushi._emit_error(
					_("tekka could not connect to maki."),
					_("Please check whether maki is running.\n"
					  "The following error occurred: %(error)s") % {
							"error": message })

			return new

		gobject_attr = False

		if attr[:2] == "g_":
			attr = attr[2:]
			gobject_attr = True

		if (attr[0] == "_" or gobject_attr
		or attr in ("connected", "remote")):
			# resolve it by gobject.__getattribute__. This function
			# will call getattr as well if there is no matching
			# method in the gobject hierarchy.
			return gobject.GObject.__getattribute__(self, attr)
		else:
			if not self._sushi:
				# return a dummy which reports an error
				return dummy
			else:
				try:
					# try local methods
					return self._sushi.__getattribute__(attr)
				except AttributeError:
					# try proxy methods
					try:
						def attr_dummy(*args, **kwargs):
							""" wrapped call to dbus proxy attr """
							try:
								return self._sushi.__getattr__(attr)(
											*args,
											**kwargs)
							except dbus.DBusException as e:
								self._emit_error(
									_("Communication error with maki."),
									_("There was an error while executing "
									  "<b>'%s</b>' with DBus: \n<b>%s</b>\n"
									  "You should keep safe that maki is "
									  "running " % (attr, e)))

						return attr_dummy
					except dbus.DBusException as e:
						# method not found, return dummy
						return errordummy(str(e))

		raise AttributeError(attr)


	# Properties
	remote = property(lambda s: s._remote, _set_remote)
	connected = property(lambda s: s._connected, _set_connected)


gobject.signal_new ("maki-connected", SushiWrapper,
	gobject.SIGNAL_ACTION, None, ())
gobject.signal_new ("maki-disconnected", SushiWrapper,
	gobject.SIGNAL_ACTION, None, ())
gobject.signal_new ("sushi-error", SushiWrapper,
	gobject.SIGNAL_ACTION, None, (str, str))


dbus_loop = DBusGMainLoop()
required_version = (1, 2, 0)
bus = None

sushi = SushiWrapper(None)

myNick = {}

_shutdown_callback = None
_nick_callback = None


def disable_sushi_on_fail(cmethod):
	""" decorator: disable sushi wrapper if connect fails """
	def new(*args, **kwargs):
		global sushi
		ret = cmethod(*args, **kwargs)
		if not ret:
			sushi._set_interface(None)
		return ret
	return new


@disable_sushi_on_fail
def connect():
	"""
	Connect to maki over DBus.
	Returns True if the connection attempt was succesful.
	If the attempt was successful, the sushi object's
	attribute "connected" is set to "True" and the object
	has more attributes through the dbus proxy so you
	can call dbus methods directly.
	"""
	global sushi, _shutdown_callback, _nick_callback, bus

	bus_address = os.getenv("SUSHI_REMOTE_BUS_ADDRESS")

	def bus_remote_error(exception):
		sushi._emit_error(
			_("tekka could not connect to maki."),
			_("Please check whether maki is running.\n"
			"The following error occurred: %(error)s") % {
				"error": str(exception) })

	def connect_session_bus():
		global bus, dbus_loop
		try:
			return dbus.SessionBus(mainloop=dbus_loop)
		except DBusException as e:
			bus_remote_error(e)
			return None

	if bus_address:
		try:
			bus = dbus.connection.Connection(bus_address,
					mainloop=dbus_loop)
		except dbus.DBusException as e:
			bus_remote_error(e)
			return False

	else:
		bus = connect_session_bus()

		if bus == None:
			return False

	if type(bus) == dbus.connection.Connection:
		sushi.remote = True

	try:
		proxy = bus.get_object("de.ikkoku.sushi", "/de/ikkoku/sushi")
	except dbus.exceptions.DBusException as e:
		bus_remote_error(e)
		return False

	sushi._set_interface(dbus.Interface(proxy, "de.ikkoku.sushi"))

	version = tuple([int(v) for v in sushi.version()])

	if not version or version < required_version:
		version_string = ".".join([str(x) for x in required_version])

		sushi._emit_error(
			_("tekka requires a newer maki version."),
			_("Please update maki to at least version %(version)s.") % {
					"version": version_string })

		sushi._set_interface(None)
		return False

	_shutdown_callback = sushi.connect_to_signal("shutdown", _shutdownSignal)
	_nick_callback = sushi.connect_to_signal("nick", _nickSignal)

	for server in sushi.servers():
		fetch_own_nick(server)

	return True


def disconnect():
	global sushi, _shutdown_callback, _nick_callback
	sushi._set_interface(None)

	if _shutdown_callback:
		_shutdown_callback.remove()

	if _nick_callback:
		_nick_callback.remove()


def parse_from (from_str):
	h = from_str.split("!", 2)

	if len(h) < 2:
		return (h[0],)

	t = h[1].split("@", 2)

	if len(t) < 2:
		return (h[0],)

	return (h[0], t[0], t[1])



def _shutdownSignal(time):
	disconnect()


def _nickSignal(time, server, from_str, new_nick):
	nick = parse_from(from_str)[0]

	if not nick or nick == get_own_nick(server):
		cache_own_nick(server, new_nick)



def sendMessage(server, channel, text):
	""" sends a PRIVMSG to channel @channel on server @server """
	text = re.sub('(^|\s)(_\S+_)(\s|$)', '\\1' + chr(31) + '\\2' + chr(31) + '\\3', text)
	text = re.sub('(^|\s)(\*\S+\*)(\s|$)', '\\1' + chr(2) + '\\2' + chr(2) + '\\3', text)

	sushi.message(server, channel, text)

# fetches the own nick for server @server from maki
def fetch_own_nick(server):
	from_str = sushi.user_from(server, "")
	nick = parse_from(from_str)[0]
	cache_own_nick(server, nick)

# caches the nick @nickname for server @server.
def cache_own_nick(server, nickname):
	myNick[server] = nickname

# returns the cached nick of server @server
def get_own_nick(server):
	if myNick.has_key(server):
		return myNick[server]
	return ""



#!/usr/bin/env python
# coding: UTF-8
"""
Copyright (c) 2008 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""


""" Purpose:
	setup and usage of submodules,
	main window signal handling,
	tekka internal signal handling,
	glue of all submodules
"""

import pygtk
pygtk.require("2.0")

import sys
import traceback

import gtk # TODO: catch gtk.Warning after init with warnings module

import os
import gobject
import pango
import dbus
import webbrowser
import locale
import types as ptypes
import logging

import gettext
from gettext import gettext as _

import gui
import gui.tabs

# local modules
from . import config
from . import com
from . import signals
from . import commands
from . import plugins

from .typecheck import types

from .lib import nick_list_store
from .lib.inline_dialog import InlineMessageDialog
from .lib.welcome_window import WelcomeWindow
from .lib.general_output_buffer import GOHTMLBuffer

from .helper import tabcompletion
from .helper import markup
from .helper.URLHandler import URLHandler

from .menus import *

import gui.builder
import gui.shortcuts

"""
Tekka intern signals
"""

def sushi_error_cb(sushi, title, message):
	""" Error in sushi interface occured. """

	def response_cb(d, i):
		gui.status.unset(title)
		d.destroy()

	d = InlineMessageDialog(title, message)
	d.connect("response", response_cb)
	gui.mgmt.show_inline_dialog(d)

	gui.status.set_visible(title, title)


def maki_connect_callback(sushi):
	""" connection to maki etablished """

	gui.mgmt.set_useable(True)


def maki_disconnect_callback(sushi):
	""" connection to maki lost """
	gui.mgmt.set_useable(False)


def tekka_server_new_nick_cb(tab, nick):
	""" New nick for the given tab. Apply the new nick in
		the GUI if the tab or a tab with the same server is active.
	"""

	activeTabs = gui.tabs.get_current_tabs()

	if (tab in activeTabs
	or (not tab.is_server() and tab.server in activeTabs)):
		gui.mgmt.set_nick(nick)


def tekka_tab_new_markup_cb(tab):
	""" Push the CellRenderer to re-render the serverTree """

	if not tab.path:
		return

	store = gui.widgets.get_object("tab_store")
	store[tab.path][0] = tab


def tekka_tab_new_message_cb(tab, mtype):
	""" A new message of the given type was received.
		If the tab is active, reset the message buffer
		and scroll the tab's textview to bottom if
		auto scrolling is enabled for this window.
	"""

	if tab.is_active():
		tab.set_new_message(None)

		if tab.window.auto_scroll and mtype:
			if tab.window.textview.is_smooth_scrolling():
				tab.window.textview.stop_scrolling()
				tab.window.textview.scroll_to_bottom(no_smooth = True)
			else:
				tab.window.textview.scroll_to_bottom()

	else:
		pass


def tekka_tab_new_name_cb(tab, name):
	tekka_tab_new_markup_cb(tab)


def tekka_tab_server_connected_cb(tab, connected):
	""" the server of the tab connected/disconnected """

	if tab.is_active():
		tab.set_useable(connected)


def tekka_channel_joined_cb(tab, switch):
	""" channel received a change on joined attribute """

	if tab.is_active():
		tab.set_useable(switch)


def tekka_tab_switched_cb(old, new):
	""" switched from tab old to tab new """

	inputBar = gui.widgets.get_object("input_entry")

	if old:
		itext = inputBar.get_text()
		old.set_input_text(itext)
		old.window.textview.set_read_line()

	inputBar.set_text("")
	inputBar.set_position(1)

	if new:
		inputBar.set_text(new.get_input_text())
		inputBar.set_position(len(inputBar.get_text()))

		if new.window.auto_scroll:
			# XXX: Needs testing!
			def check_for_scrolling():
				sw = new.window
				adj = sw.get_vadjustment()

				if adj.get_value() != (adj.upper - adj.page_size):
					sw.textview.scroll_to_bottom( no_smooth = True )
				else:
					print "No need for scrolling!"
				return False

			gobject.idle_add(check_for_scrolling)


def tekka_tab_add_cb(tab):
	""" a tab is added """

	if gui.mgmt.is_welcome_screen():
		# FIXME: this is called often if the tab is not changed
		gui.mgmt.visibility.show_welcome_screen(False)


def tekka_tab_remove_cb(tab):
	""" a tab is about to be removed """

	if gui.tabs.get_current_tab() == tab:
		# switch to another tab

		if tab.is_server():
			# server and children are removed, choose
			# another server
			server = gui.tabs.get_next_server(tab)

			if server:
				tabs = gui.tabs.get_all_tabs(servers = [server.name])
				nextTab = tabs[0]
			else:
				nextTab = None
		else:
			nextTab = gui.tabs.get_next_tab(tab)

		if None == nextTab:
			# lock interface
			# XXX:  maybe the inputBar should
			# XXX:: useable, though.
			gui.mgmt.set_useable(False)
		else:
			nextTab.switch_to()

	elif (tab.is_server()
	and len(gui.widgets.get_object("tab_store")) == 1):
		gui.mgmt.set_useable(False)


def tekka_channel_topic_changed_cb(tab, topic):
	if not tab.is_active(): return

	if (config.get_bool("tekka","hide_topic_if_empty")
	and config.get_bool("tekka", "show_topic_bar")):
		if topic:
			gui.mgmt.visibility.show_topic_bar(True)
		else:
			gui.mgmt.visibility.show_topic_bar(False)


def mainWindow_scroll_event_cb(mainWindow, event):
	""" MOD1 + SCROLL_DOWN -> Next tab
		MOD1 + SCROLL_UP -> Prev. tab
	"""

	if (event.state & gtk.gdk.MOD1_MASK
	and event.direction == gtk.gdk.SCROLL_DOWN):
		gui.tabs.switch_to_next()

	elif (event.state & gtk.gdk.MOD1_MASK
	and event.direction == gtk.gdk.SCROLL_UP):
		gui.tabs.switch_to_previous()


def mainWindow_delete_event_cb(mainWindow, event):
	"""
		If hide_on_close and the status icon are enabled,
		hide the main window. Otherwise stop the main loop.

		On hide, a read-line will be inserted in every tab.
	"""

	statusIcon = gui.widgets.get_object("status_icon")

	if (config.get_bool("tekka", "hide_on_close")
	and statusIcon and statusIcon.get_visible()):

		for tab in gui.tabs.get_all_tabs():
			tab.window.textview.set_read_line()

		mainWindow.hide()

		return True

	else:
		gtk.main_quit()


def mainWindow_focus_in_event_cb(mainWindow, event):
	"""
		Reset urgent status (if given)
	"""

	gui.mgmt.set_urgent(False)
	return False


def mainWindow_size_allocate_cb(mainWindow, alloc):
	""" Main window was resized, store the new size in the config. """

	if not mainWindow.window.get_state() & gtk.gdk.WINDOW_STATE_MAXIMIZED:
		config.set("sizes","window_width",alloc.width)
		config.set("sizes","window_height",alloc.height)


def mainWindow_window_state_event_cb(mainWindow, event):
	""" Window state was changed.
		If it's maximized or unmaximized, save that state.
	"""

	if event.new_window_state & gtk.gdk.WINDOW_STATE_MAXIMIZED:
		config.set("tekka","window_maximized","True")
	else:
		config.set("tekka","window_maximized","False")


def inputBar_activate_cb(inputBar):
	""" Enter hit, pass the inputBar text over to the
		commands.parseInput method and add the text to the input history.

		The inputBar is cleared after that.
	"""

	text = inputBar.get_text()

	tab = gui.tabs.get_current_tab()

	inputBar.set_text("")

	commands.parseInput(text)

	if tab:
		tab.input_history.add_entry(text)
		tab.input_history.reset()


def inputBar_key_press_event_cb(inputBar, event):
	""" Up -> Input history previous entry
		Down -> Input history next entry
		Tab -> Completion of the current word

		Everything else than tab ->
			No further completion wished, tell that.
	"""

	key =  gtk.gdk.keyval_name(event.keyval)
	tab =  gui.tabs.get_current_tab()

	text = unicode(inputBar.get_text(), "UTF-8")

	if key == "Up":
		# get next input history item
		if not tab:
			return

		hist = tab.input_history.get_previous()

		if hist != None:
			inputBar.set_text(hist)
			inputBar.set_position(len(hist))

	elif key == "Down":
		# get previous input history item
		if not tab:
			return

		hist = tab.input_history.get_next()

		if hist == None:
			return

		inputBar.set_text(hist)
		inputBar.set_position(len(hist))

	elif key == "Tab":
		# tab completion comes here.

		tabcompletion.complete(tab, inputBar, text)
		return True

	if key != "Tab":
		tabcompletion.stopIteration()


def notificationWidget_remove_cb(area, widget):
	""" restore the focus if a inline dialog is closed """
	gui.widgets.get_object("input_entry").grab_focus()


def outputShell_widget_changed_cb(shell, old_widget, new_widget):
	""" old_widget: OutputWindow
		new_widget: OutputWindow

		Set the current content of the output_shell in widgets store.
		- output_window <- new_widget
		- output        <- new_widget.textview
	"""
	if (type(old_widget) == WelcomeWindow
	and type(new_widget) != WelcomeWindow):
		gui.mgmt.visibility.show_welcome_screen(False)

	gui.widgets.remove_object("output_window")
	gui.widgets.add_object(new_widget, "output_window")

	gui.widgets.remove_object("output")
	gui.widgets.add_object(new_widget.textview, "output")


def serverTree_misc_menu_reset_activate_cb(menuItem):
	""" reset the markup of all tabs """

	for tab in gui.tabs.get_all_tabs():
		tab.set_new_message(None)


def serverTree_button_press_event_cb(serverTree, event):
	"""
		A row in the server tree was activated.
		The main function of this method is to
		cache the current activated row as path.
	"""

	try:
		path = serverTree.get_path_at_pos(int(event.x),int(event.y))[0]
		tab = serverTree.get_model()[path][0]
	except Exception as e:
		tab = None

	if event.button == 1:
		# activate the tab

		if tab:
			gui.tabs.switch_to_path(path)

	elif event.button == 2:
		# if there's a tab, ask to close
		if tab:
			askToRemoveTab(tab)

	elif event.button == 3:
		# popup tab menu

		if tab:
			menu = servertree_menu.ServerTreeMenu().get_menu(tab)

			if not menu:
				logging.error("error in creating server tree tab menu.")
				return False

			else:
				menu.popup(None, None, None, event.button, event.time)
				return True

		else:
			# display misc. menu
			menu = gtk.Menu()
			reset = gtk.MenuItem(label=_(u"Reset markup"))
			reset.connect("activate",
				serverTree_misc_menu_reset_activate_cb)
			menu.append(reset)
			reset.show()
			menu.popup(None,None,None,event.button,event.time)

	return False


def serverTree_row_activated_cb(serverTree, path, column):
	""" open the history dialog for the pointed tab """
	model = serverTree.get_model()
	tab = model[path][0]

	# don't show the history dialog for server tabs, they don't
	# have a history.
	if type(tab) != gui.tabs.TekkaServer:
		gui.dialogs.show_dialog("history", tab)


def nickList_row_activated_cb(nickList, path, column):
	"""
		The user activated a nick in the list.

		If there's a nick in the row a query
		for the nick on the current server will be opened.
	"""
	serverTab,channelTab = gui.tabs.get_current_tabs()

	try:
		name = nickList.get_model()[path][nick_list_store.COLUMN_NICK]
	except TypeError:
		# nickList has no model
		return
	except IndexError:
		# path is invalid
		return

	if gui.tabs.search_tab(serverTab.name, name):
		# already a query open
		return

	query = gui.tabs.create_query(serverTab, name)
	query.connected = True

	gui.tabs.add_tab(serverTab, query)

	query.print_last_log()
	query.switch_to()


def nickList_button_press_event_cb(nickList, event):
	"""
		A button pressed inner nickList.

		If it's the right mouse button and there
		is a nick at the coordinates, pop up a menu
		for setting nick options.
	"""
	if event.button == 3:
		# right mouse button pressed.

		path = nickList.get_path_at_pos(int(event.x), int(event.y))

		nick = None

		# get marked nick
		try:
			nick = nickList.get_model()[path[0]]
		except TypeError:
			# no model
			pass
		except IndexError:
			# path is "invalid"
			pass

		if nick:
			# display nick specific menu

			nick = nick[nick_list_store.COLUMN_NICK]

			menu = nicklist_menu.NickListMenu().get_menu(nick)

			if not menu:
				return False

			# finaly popup the menu
			menu.popup(None, None, None, event.button, event.time)

	return False


def nicks_view_query_tooltip_cb(view, x, y, kbdmode, tooltip):
	""" generate a tooltip with the awaymessage of the
		nick at the given x/y coordinates.
	"""

	# TODO: would be nice to have ident string of the nick here

	cursor = view.get_path_at_pos(x, y)

	if not cursor:
		return

	user_row = view.get_model()[cursor[0]]
	tip = ""

	# away message appendix
	if user_row[nick_list_store.COLUMN_AWAY]:
		# the user is away

		(server,_) = gui.tabs.get_current_tabs()


		if server:

			"""
			msg = com.sushi.awaymessage(server.name,
								user_row[nick_list_store.COLUMN_NICK])
			"""
			# TODO: retrieve awaymessage
			pass



""" Shortcut callbacks """

def inputBar_shortcut_ctrl_u(inputBar, shortcut):
	""" Ctrl + U was hit, clear the inputBar """

	gui.widgets.get_object("input_entry").set_text("")


def output_shortcut_ctrl_l(inputBar, shortcut):
	"""
		Ctrl+L was hit, clear the outputs.
	"""
	gui.mgmt.clear_all_outputs()


def output_shortcut_ctrl_f(inputBar, shortcut):
	""" show/hide the search toolbar """
	sb = gui.widgets.get_object("output_searchbar")

	if sb.get_property("visible"):
		sb.hide()
		gui.widgets.get_object("input_entry").grab_focus()
	else:
		sb.show_all()
		sb.grab_focus()


def output_shortcut_ctrl_g(inputBar, shortcut):
	""" search further """

	gui.widgets.get_object("output_searchbar").search_further()


def serverTree_shortcut_ctrl_Page_Up(serverTree, shortcut):
	""" Ctrl+Page_Up was hit, go up in server tree """

	gui.tabs.switch_to_previous()


def serverTree_shortcut_ctrl_Page_Down(serverTree, shortcut):
	""" Ctrl+Page_Down was hit, go down in server tree """

	gui.tabs.switch_to_next()


def askToRemoveTab(tab):
	def response_handler(dialog, response_id):

		if response_id == gtk.RESPONSE_YES:

			if tab.is_channel():
				com.sushi.part(tab.server.name, tab.name,
					config.get("chatting", "part_message", ""))

			elif tab.is_server():
				com.sushi.quit(tab.name,
					config.get("chatting", "quit_message", ""))

			gui.tabs.remove_tab(tab)

		dialog.destroy()

	if tab.is_channel():
		message = _(u"Do you really want to close channel “%(name)s”?")
	elif tab.is_query():
		message = _(u"Do you really want to close query “%(name)s”?")
	elif tab.is_server():
		message = _(u"Do you really want to close server “%(name)s”?")

	dialog = InlineMessageDialog(
		message % { "name": tab.name },
		icon=gtk.STOCK_DIALOG_QUESTION,
		buttons=gtk.BUTTONS_YES_NO
	)
	dialog.connect("response", response_handler)

	gui.mgmt.show_inline_dialog(dialog)


def serverTree_shortcut_ctrl_w(serverTree, shortcut):
	""" Ctrl+W was hit, close the current tab (if any) """

	tab = gui.tabs.get_current_tab()

	if not tab:
		return

	askToRemoveTab(tab)


def output_shortcut_Page_Up(inputBar, shortcut):
	"""
		Page_Up was hit, scroll up in output
	"""
	vadj = gui.widgets.get_object("output_window").get_vadjustment()

	if vadj.get_value() == 0.0:
		return # at top already

	n = vadj.get_value()-vadj.page_size
	if n < 0: n = 0
	gobject.idle_add(vadj.set_value,n)


def output_shortcut_Page_Down(inputBar, shortcut):
	""" Page_Down was hit, scroll down in output """

	vadj = gui.widgets.get_object("output_window").get_vadjustment()

	if (vadj.upper - vadj.page_size) == vadj.get_value():
		return # we are already at bottom

	n = vadj.get_value()+vadj.page_size

	if n > (vadj.upper - vadj.page_size):
		n = vadj.upper - vadj.page_size

	gobject.idle_add(vadj.set_value,n)


def inputBar_shortcut_ctrl_c(inputBar, shortcut):
	"""
		Ctrl + C was hit.
		Check every text input widget for selection
		and copy the selection to clipboard.
		FIXME: this solution sucks ass.
	"""
	buffer = gui.widgets.get_object("output").get_buffer()
	goBuffer = gui.widgets.get_object("general_output").get_buffer()
	topicBar = gui.widgets.get_object("topic_label")
	cb = gtk.Clipboard()

	if buffer.get_property("has-selection"):
		buffer.copy_clipboard(cb)
	elif inputBar.get_selection_bounds():
		inputBar.copy_clipboard()
	elif goBuffer.get_property("has-selection"):
		goBuffer.copy_clipboard(cb)
	elif topicBar.get_selection_bounds():
		bounds = topicBar.get_selection_bounds()
		text = unicode(topicBar.get_text(), "UTF-8")
		text = text[bounds[0]:bounds[1]]
		cb.set_text(text)


def changeTopic_shortcut(inputBar, shortcut):
	""" The user wants to change the topic for the current tab.
	"""
	channel = gui.tabs.get_current_tab()
	if not channel or not channel.is_channel():
		return

	menu = servertree_menu.ServerTreeMenu()
	menu.get_menu(channel)
	menu.widgets.get_object("setTopicItem").activate()


def serverTree_query_tooltip_cb(widget, x, y, kbdmode, tooltip):
	""" show tooltips for treeview rows.

		Server tabs:
			Nick: <nickname>

		Channel tabs:
			Users: <count>
			Topic: <topic>
			Last Sentence: <last sentence>

		Query tabs:
			Last Sentence: <last sentence>
	"""

	def limit(s):
		limit = int(config.get("tekka","popup_line_limit"))
		if len(s) > limit:
			return markup.escape(s[:limit-3]+u"...")
		return markup.escape(s)

	path = widget.get_path_at_pos(x,y)

	if not path:
		return

	path = path[0]

	try:
		tab = widget.get_model()[path][0]
	except IndexError:
		return

	if tab.is_server():
		# TODO: away status
		s = "<b>" + _("Nickname: ") + "</b>" +  markup.escape(tab.nick)

	elif tab.is_channel():
		s = "<b>" +_("User: ") + "</b>" + str(len(tab.nickList)) +\
			"\n<b>" + _("Topic: ") + "</b>" +\
				limit(tab.topic) +\
			"\n<b>" + _("Last sentence: ") + "</b>" +\
				limit(tab.window.textview.get_last_line())

	elif tab.is_query():
		s = "<b>" + _("Last sentence: ") + "</b>" +\
			limit(tab.window.textview.get_last_line())

	tooltip.set_markup(s)

	return True


def serverTree_render_server_cb(column, renderer, model, iter):
	""" Renderer func for column "Server" in servertree """

	tab = model.get(iter, 0)

	if not tab or not tab[0]:
		return

	renderer.set_property("markup",tab[0].markup())


def nickList_render_nicks_cb(column, renderer, model, iter):
	""" Renderer func for column "Nicks" in NickList """

	if not com.sushi.connected:
		# do not render if no connection exists
		return

	# highlight own nick
	serverTab = gui.tabs.get_current_tabs()[0]

	if not serverTab:
		return

	nick = model.get(iter, 1)
	away = model.get(iter, 2)

	if not nick:
		return

	nick = nick[0]
	away = away[0]

	# highlight own nick
	if com.get_own_nick(serverTab.name) == nick:
		renderer.set_property("weight", pango.WEIGHT_BOLD)
	else:
		renderer.set_property("weight", pango.WEIGHT_NORMAL)

	if away:
		renderer.set_property("style", pango.STYLE_ITALIC)
	else:
		renderer.set_property("style", pango.STYLE_NORMAL)



def treemodel_rows_reordered_cb(treemodel, path, iter, new_order):
	""" new_order is not accessible, so hack arround it... """

	# explicit import because what we do is bad.
	# there should be no one writing on current_path
	from gui.tabs.current import _set_current_path

	updated = False
	for row in treemodel:
		if not row[0]:
			continue

		if gui.tabs.get_current_path() == row[0].path and not updated:
			# update the currentPath cache
			_set_current_path(row.path)
			updated = True

		# update the tab's path cache
		row[0].path = row.path

		for child in row.iterchildren():
			if not child[0]:
				continue

			if (gui.tabs.get_current_path() == child[0].path
			and not updated):
				_set_current_path(child.path)
				updated = True

			# update path's tab cache
			child[0].path = child.path


def paned_notify_cb(paned, gparam):
	if not paned_notify_cb.init_done:
		return

	if gparam.name == "position":

		# only save if there are no inline dialogs displayed
		ids = gui.widgets.get_object("notification_vbox").get_children()
		if len(ids) == 0:
			config.set("sizes", gtk.Buildable.get_name(paned),
						paned.get_property("position"))

paned_notify_cb.init_done = False


""" Initial setup routines """

def setup_main_window():
	"""
		- set window title
		- set window icon
		- set window size
		- set window state
	"""
	win = gui.widgets.get_object("main_window")

	win.set_title("tekka IRC client")

	if config.get_bool("tekka", "rgba"):
		colormap = win.get_screen().get_rgba_colormap()
		if colormap:
			gtk.widget_set_default_colormap(colormap)

	try:
		img = gtk.icon_theme_get_default().load_icon("tekka",64,0)

		win.set_icon(img)
	except gobject.GError:
		# file not found
		pass

	# Restore sizes from last start
	width = config.get("sizes","window_width")
	height = config.get("sizes","window_height")

	if width and height:
		win.resize(int(width),int(height))


	# Restore window state from last start
	if config.get_bool("tekka","window_maximized"):
		win.maximize()

	# enable scrolling through server tree by scroll wheel
	def kill_mod1_scroll(w,e):
		if e.state & gtk.gdk.MOD1_MASK:
			w.emit_stop_by_name("scroll-event")

	for widget in ("general_output_window",
				   "tabs_window",
				   "nicks_window"):
		gui.widgets.get_object(widget).connect("scroll-event",
											   kill_mod1_scroll)

	win.show()


def setup_tabs_view():
	""" Setup tab sorting, setup tab rendering """
	model = gui.widgets.get_object("tab_store")

	# Sorting
	def cmpl(m,i1,i2):
		" compare columns lower case "

		a = m.get_value(i1, 0)
		b = m.get_value(i2, 0)

		c, d = None, None

		if a: c=a.name.lower()
		if b: d=b.name.lower()
		return cmp(c,d)

	model.set_sort_func(1, lambda m,i1,i2,*x: cmpl(m,i1,i2))
	model.set_sort_column_id(1, gtk.SORT_ASCENDING)

	# Setup the renderer
	column = gui.widgets.get_object("tabs_view_name_column")
	column.set_cell_data_func(
				gui.widgets.get_object("tabs_view_name_renderer"),
				serverTree_render_server_cb)


def setup_general_ouptut():
	""" set the textview's buffer to a GOHTMLBuffer and add
		the textview as general_output to the widgets store
	"""
	w = gui.widgets.get_object("general_output_window")
	w.textview.set_buffer(GOHTMLBuffer(handler=URLHandler))
	gui.widgets.add_object(w.textview, "general_output")


def setup_nicks_view():
	""" setup custom rendering of nick column """
	column = gui.widgets.get_object("nicks_store_nick_column")
	column.set_cell_data_func(
			gui.widgets.get_object("nicks_store_nick_renderer"),
			nickList_render_nicks_cb)


def load_paned_positions():
	""" restore the positions of the
		paned dividers for the list,
		main and output paneds.
	"""

	paneds = [
		"list_vpaned",
		"main_hpaned",
		"output_vpaned"]

	for paned_name in paneds:
		paned = gui.widgets.get_object(paned_name)

		position = config.get("sizes", paned_name, None)
		paned.set_property("position-set", True)

		if position == None:
			continue

		try:
			paned.set_position(int(position))
		except ValueError:
			logging.error("Failed to set position for paned %s" % (
				paned.name))
			continue


def setup_paneds():
	load_paned_positions()
	paned_notify_cb.init_done = True
	return False


def setup_fonts():
	""" add default font callback """
	try:
		import gconf

		def default_font_cb (client, id, entry, data):
			if not config.get_bool("tekka", "use_default_font"):
				return

			gui.mgmt.apply_new_font()

		c = gconf.client_get_default()

		c.add_dir("/desktop/gnome/interface", gconf.CLIENT_PRELOAD_NONE)
		c.notify_add("/desktop/gnome/interface/monospace_font_name",
			default_font_cb)

	except:
		# ImportError or gconf reported a missing dir.
		pass


def setup_topic_label():
	""" tooltip style box arround topic label """

	def expose_event_cb(box, event):
		a = box.get_allocation()

		box.style.paint_flat_box(
			box.window,
			gtk.STATE_NORMAL,
			gtk.SHADOW_ETCHED_IN,
			None,
			box,
			"tooltip",
			a.x,
			a.y,
			a.width,
			a.height - 1
		)
		return False
	gui.widgets.get_object("topic_label").connect(
		"expose-event", expose_event_cb)



def setupGTK():
	""" Set locale, load UI file, connect signals, setup widgets. """

	uifiles = config.get("uifiles", default={})

	# setup locale stuff
	try:
		locale.setlocale(locale.LC_ALL, '')
		locale.bindtextdomain("tekka", config.get("tekka","locale_dir"))
		locale.textdomain("tekka")
	except:
		pass

	# Fix about dialog URLs
	def about_dialog_url_hook (dialog, link, data):
		if gtk.gtk_version >= (2, 16, 0): return
		webbrowser.open(link)
	gtk.about_dialog_set_url_hook(about_dialog_url_hook, None)

	# Fire gettext up with our locale directory
	gettext.bindtextdomain("tekka", config.get("tekka","locale_dir"))
	gettext.textdomain("tekka")

	# parse ui file for main window
	gui.builder.load_main_window(uifiles["mainwindow"])

	setup_main_window()


	mmc = gui.widgets.get_object("main_menu_context")

	# tell the searchbar where it can get it's input
	gui.widgets.get_object("output_searchbar").textview_callback =\
		lambda: gui.widgets.get_object("output")

	# connect tab control signals
	gui.tabs.add_callbacks({
		"new_message": tekka_tab_new_message_cb,
		"new_name": tekka_tab_new_name_cb,
		"add": tekka_tab_add_cb,
		"remove": tekka_tab_remove_cb,
		"new_markup": tekka_tab_new_markup_cb,
		"server_connected": tekka_tab_server_connected_cb,
		"joined": tekka_channel_joined_cb,
		"new_nick": tekka_server_new_nick_cb,
		"tab_switched": tekka_tab_switched_cb,
		"topic": tekka_channel_topic_changed_cb,
		})

	# connect main window signals:
	sigdic = {
		# main window signals
		"main_window_delete_event":
			mainWindow_delete_event_cb,
		"main_window_focus_in_event":
			mainWindow_focus_in_event_cb,
		"main_window_size_allocate":
			mainWindow_size_allocate_cb,
		"main_window_window_state_event":
			mainWindow_window_state_event_cb,
		"main_window_scroll_event":
			mainWindow_scroll_event_cb,

		# server tree signals
		"tabs_view_button_press_event" :
			serverTree_button_press_event_cb,
		"tabs_view_row_activated":
			serverTree_row_activated_cb,
		"tabs_view_query_tooltip":
			serverTree_query_tooltip_cb,

		# Store of the tabs view
		"tab_store_rows_reordered":
			treemodel_rows_reordered_cb,

		# Input entry...
		"input_entry_activate":
			inputBar_activate_cb,
		"input_entry_key_press_event":
			inputBar_key_press_event_cb,

		# Notification VBox
		"notification_vbox_remove":
			notificationWidget_remove_cb,

		"output_shell_widget_changed":
			outputShell_widget_changed_cb,

		# nick list signals
		"nicks_view_row_activated":
			nickList_row_activated_cb,
		"nicks_view_button_press_event":
			nickList_button_press_event_cb,
		"nicks_view_query_tooltip":
			nicks_view_query_tooltip_cb,

		# watch for position change of paneds
		"list_vpaned_notify":
			paned_notify_cb,
		"main_hpaned_notify":
			paned_notify_cb,
		"output_vpaned_notify":
			paned_notify_cb,

	# tekka menu context
		"tekka_server_list_item_activate":
			mmc.tekka.connect_activate_cb,
		"tekka_quit_item_activate":
			mmc.tekka.quit_activate_cb,
	# maki menu context
		"maki_connect_item_activate":
			mmc.maki.connect_activate_cb,
		"maki_disconnect_item_activate":
			mmc.maki.disconnect_activate_cb,
		"maki_shutdown_item_activate":
			mmc.maki.shutdown_activate_cb,
	# view menu context
		"view_general_output_item_toggled":
			mmc.view.showGeneralOutput_toggled_cb,
		"view_side_pane_item_toggled":
			mmc.view.showSidePane_toggled_cb,
		"view_status_bar_item_toggled":
			mmc.view.showStatusBar_toggled_cb,
		"view_status_icon_item_toggled":
			mmc.view.showStatusIcon_toggled_cb,
		"view_topic_bar_item_toggled":
			mmc.view.showTopicBar_toggled_cb,
	# tools menu context
		"tools_channel_list_item_activate":
			mmc.tools.channelList_activate_cb,
		"tools_file_transfers_item_activate" :
			mmc.tools.dcc_activate_cb,
		"tools_plugins_item_activate" :
			mmc.tools.plugins_activate_cb,
		"tools_debug_item_activate" :
			mmc.tools.debug_activate_cb,
		"tools_preferences_item_activate" :
			mmc.tools.preferences_activate_cb,
	# help menu context
		"help_irc_colors_item_activate":
			mmc.help.colors_activate_cb,
		"help_about_item_activate":
			mmc.help.about_activate_cb,
	}

	gui.widgets.connect_signals(sigdic)

	# push status messages directly in the status bar
	gui.status.connect("set-visible-status",
		lambda w,s,m: gui.widgets.get_object("statusbar")\
		.push(gui.status.id(s), m))

	# pop status message if they're unset
	gui.status.connect("unset-status",
		lambda w,s: gui.widgets.get_object("statusbar")\
		.pop(gui.status.id(s)))

	# initialize output_shell again (signals are connected now)
	gui.widgets.get_object("output_shell").reset()

	# setup more complex widgets
	setup_tabs_view()
	setup_nicks_view()
	setup_general_ouptut()
	setup_topic_label()

	# apply visibility to widgets from config
	mmc.view.apply_visibility_settings()
	gui.mgmt.visibility.apply_visibility_from_config()

	setup_fonts()

	# set input font
	gui.mgmt.set_font(gui.widgets.get_object("input_entry"),
					  gui.mgmt.get_font())

	# set general output font
	gui.mgmt.set_font(gui.widgets.get_object("general_output"),
					  gui.mgmt.get_font())

	gui.shortcuts.add_handlers({
			"clear_outputs": output_shortcut_ctrl_l,
			"output_page_up": output_shortcut_Page_Up,
			"output_page_down": output_shortcut_Page_Down,
			"input_clear_line": inputBar_shortcut_ctrl_u,
			"input_search": output_shortcut_ctrl_f,
			"input_search_further": output_shortcut_ctrl_g,
			"input_copy": inputBar_shortcut_ctrl_c,
			"change_topic": changeTopic_shortcut,
			"servertree_previous": serverTree_shortcut_ctrl_Page_Up,
			"servertree_next": serverTree_shortcut_ctrl_Page_Down,
			"servertree_close": serverTree_shortcut_ctrl_w,
			"show_sidepane": lambda w,s: w.set_active(not w.get_active()),
		})
	gui.shortcuts.setup_shortcuts()

	gui.mgmt.visibility.show_welcome_screen(True)

	# disable the GUI and wait for commands :-)
	gui.mgmt.set_useable(False)

	gobject.idle_add(setup_paneds)


def tekka_excepthook(extype, exobj, extb):
	""" we got an exception, print it in a dialog box and,
		if possible, to the standard output.
	"""

	message = "%s\n%s: %s\n" % (
		"".join(traceback.format_tb(extb)),
		extype.__name__,
		str(exobj))

	try:
		print >> sys.stderr, message
	except:
		pass

	self = tekka_excepthook

	def dialog_response_cb(dialog, rid):
		del self.dialog
		dialog.destroy()

	if not hasattr(self, "dialog"):
		from .lib.error_dialog import ErrorDialog

		self.dialog = ErrorDialog(message)
		self.dialog.connect("response", dialog_response_cb)
		self.dialog.show_all()
	else:
		self.dialog.set_message(message)

	sys.__excepthook__(extype, exobj, extb)


def setup_logging():
	""" set the path of the logfile to tekka.logfile config
		value and create it (including path) if needed.
		After that, add a logging handler for exceptions
		which reports exceptions catched by the logger
		to the tekka_excepthook. (DBus uses this)
	"""
	try:
		class ExceptionHandler(logging.Handler):
			""" handler for exceptions caught with logging.error.
				dump those exceptions to the exception handler.
			"""
			def emit(self, record):
				if record.exc_info:
					tekka_excepthook(*record.exc_info)

		logfile = config.get("tekka","logfile")
		logdir = os.path.dirname(logfile)

		if not os.path.exists(logdir):
			os.makedirs(logdir)

		logging.basicConfig(filename = logfile, level = logging.DEBUG,
			filemode="w")

		logging.getLogger("").addHandler(ExceptionHandler())

	except BaseException as e:
		print >> sys.stderr, "Logging init error: %s" % (e)


def setup():
	""" Setup the UI """

	# load config file, apply defaults
	config.setup()

	# create logfile, setup logging module
	setup_logging()

	# setup callbacks
	com.sushi.g_connect("sushi-error", sushi_error_cb)
	com.sushi.g_connect("maki-connected", maki_connect_callback)
	com.sushi.g_connect("maki-disconnected", maki_disconnect_callback)
	signals.setup()

	# build graphical interface
	setupGTK()

	# setup exception handler
	sys.excepthook = tekka_excepthook


def main(args):
	""" Fire up the UI """

	# connect to maki daemon
	com.connect()

	plugins.load_autoloads()

	# start main loop
	gtk.main()

	# after main loop break, write config
	config.write_config_file()

	# At last, close maki if requested
	if config.get_bool("tekka", "close_maki_on_close"):
		com.sushi.shutdown(config.get("chatting", "quit_message", ""))




"""
Copyright (c) 2009-2010 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gtk
import gobject

from .output_window import OutputWindow
from ..typecheck import types

class OutputShell(gtk.VBox):

	""" A shell for one OutputWindow with
		methods to display another OutputWindow.
	"""

	__gtype_name__ = "OutputShell"

	@types (widget = OutputWindow)
	def __init__(self, window=OutputWindow()):
		""" Takes a default window which is shown if reset() is
			called (which is the default).
		"""
		gtk.VBox.__init__(self)

		self.init_window = window
		self.output_window = None

		self.reset()

	@types (new_window = OutputWindow)
	def set(self, new_window):
		""" Set a new OutputWindow which replaces
			the current.

			Emits widget-changed with the old and the
			new widget.
		"""
		old_window = self.output_window

		if old_window:
			self.remove(old_window)

		self.pack_start(new_window)
		self.output_window = new_window

		self.emit("widget-changed", old_window, new_window)

	def reset(self):
		""" Reset to the default window. """
		self.set(self.init_window)

	def get(self):
		""" Return the current OutputWindow """
		return self.output_window

gobject.signal_new(
	"widget-changed", OutputShell,
	gobject.SIGNAL_ACTION, gobject.TYPE_NONE,
	(gobject.TYPE_PYOBJECT,gobject.TYPE_PYOBJECT))




"""
Copyright (c) 2009-2010 Marian Tietz
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

import gtk
import gobject

from .output_textview import OutputTextView
from .. import config

from math import ceil, floor

class OutputWindow(gtk.ScrolledWindow):

	""" A gtk.ScrolledWindow with a TextView inside of it.

		This widget watches for autoscrolling and
		adjusts the scrollbar on size-allocations.

		This widget is supposed to be hold by a OutputShell.
	"""

	__gtype_name__ = "OutputWindow"

	# overload show
	def do_show(self):
		gtk.ScrolledWindow.do_show(self)
		self.textview.show()

	def __init__(self):
		gtk.ScrolledWindow.__init__(self)

		self.set_properties(
			hscrollbar_policy = gtk.POLICY_AUTOMATIC,
			vscrollbar_policy = gtk.POLICY_AUTOMATIC,
				  shadow_type = gtk.SHADOW_ETCHED_IN )

		self.textview = OutputTextView()
		self.auto_scroll = True

		self.add(self.textview)

		self.old_allocation = self.get_allocation()

		# XXX: redundant code, see main.py::setup_mainWindow
		def kill_mod1_scroll_cb(w,e):
			if e.state & gtk.gdk.MOD1_MASK:
				w.emit_stop_by_name("scroll-event")

		self.connect("scroll-event", kill_mod1_scroll_cb)

		def size_allocate_cb(win, alloc):
			""" Called when the window has a new size.
				If the new size differs from the old size,
				determine if we wanted to be at the bottom
				(auto_scroll = True) and scroll down.
			"""
			if alloc.height != self.old_allocation.height:

				if self.auto_scroll:

					def doit():
						self.textview.scroll_to_bottom(no_smooth = True)
						return False

					gobject.idle_add(doit)

			self.old_allocation = alloc

		self.connect("size-allocate", size_allocate_cb)

		def value_changed_cb(sbar):
			""" Called if the scrollbar value has changed.

				If the scrollbar is at the bottom,
				set auto_scroll to True.

				If we're in the middle of a smooth scrolling
				action, and self.auto_scroll is True, it will
				be True after all.

				In all other cases, we don't want auto scroll
				anymore.
			"""

			def idle_handler_cb():
				adjust = sbar.get_property("adjustment")

				line_padding = floor(int(config.get("tekka", "line_spacing")) / 2)
				end_position = adjust.upper - adjust.page_size - line_padding
				cur_position = sbar.get_value()

				if end_position == 0 or cur_position == 0:
				    miss = 1
				else:
				    miss = 1 - min(end_position, cur_position) / max(end_position, cur_position)

				if (self.auto_scroll and self.textview.is_smooth_scrolling()):
					# XXX: instead of setting, ignore this completely.
					self.auto_scroll = True
				elif miss < 0.01:
					self.auto_scroll = True
				else:
					self.auto_scroll = False

				return False

			# XXX:  maybe one can get rid of this if using connect_after
			# XXX:: instead of connect
			gobject.idle_add(idle_handler_cb)

		def doit():
			self.get_vscrollbar().connect("value-changed", value_changed_cb)
		gobject.idle_add(doit)




""" beep on highlight """

import sushi
import string

def do_beep():
	""" tries several beep methods for
		different platforms and frameworks
		and returns if one method was
		successful.
	"""
	try:
		import winsound
		return winsound.Beep(30, 1)
	except:
		pass

	try:
		import MacOS
		# not in 64 bit
		return MacOS.SysBeep()
	except:
		pass

	try:
		import gtk
		return gtk.gdk.beep()
	except:
		pass

	try:
		tty = file("/dev/tty", "w")
		tty.write(chr(7))
		tty.close()
		return None
	except:
		pass

plugin_info = ("Let the PC speaker beep on a highlight.",
	"1.0", "Marian Tietz")
plugin_options = ()

class beep (sushi.Plugin):

	def __init__(self):
		sushi.Plugin.__init__(self, "beep")

		self.get_bus().connect_to_signal("message",
			self.message_cb)

	def message_cb(self, time, server, usr, target, text):

		def has_highlight(text, needle):
			punctuation = string.punctuation + " \n\t"
			ln = len(needle)
			for line in text.split("\n"):
				line = line.lower()
				i = line.find(needle)
				if i >= 0:
					if (line[i-1:i] in punctuation
					and line[ln+i:ln+i+1] in punctuation):
						return True
			return False

		nick = self.get_nick(server).lower()
		target = target.lower()

		if nick == target or has_highlight(text, nick):
			do_beep()
